
"""
Provides the main class for logging, :class:`Log`, and some helpers.
"""

from __future__ import print_function

import logging
import os
import sys
try:
  import StringIO
except ImportError:
  # noinspection PyPep8Naming
  import io as StringIO
import threading
from threading import RLock
import contextlib
import string
import typing

PY3 = sys.version_info[0] >= 3


class Stream:
  """
  Simple stream wrapper, which provides :func:`write` and :func:`flush`.
  """

  # noinspection PyShadowingNames
  def __init__(self, log, lvl):
    """
    :type log: logging.Logger
    :type lvl: int
    """
    self.buf = StringIO.StringIO()
    self.log = log
    self.lvl = lvl
    self.lock = RLock()

  def write(self, msg):
    """
    :param str msg:
    """
    with self.lock:
      if msg == '\n':
        self.flush()
      else:
        self.buf.write(msg)

  def flush(self):
    """
    Flush, i.e. writes to the log.
    """
    with self.lock:
      self.buf.flush()
      self.log.log(self.lvl, self.buf.getvalue())
      self.buf.truncate(0)
      # truncate does not change the current position.
      # In Python 2.7, it incorrectly does. See: http://bugs.python.org/issue30250
      self.buf.seek(0)


class Log:
  """
  The main logging class.
  """

  def __init__(self):
    self.initialized = False
    self.filename = None  # type: typing.Optional[str]
    self.v = None  # type: typing.Optional[typing.List[logging.Logger]]
    self.verbose = None  # type: typing.Optional[typing.List[bool]]
    self.error = None  # type: typing.Optional[Stream]
    self.v0 = None  # type: typing.Optional[Stream]
    self.v1 = None  # type: typing.Optional[Stream]
    self.v2 = None  # type: typing.Optional[Stream]
    self.v3 = None  # type: typing.Optional[Stream]
    self.v4 = None  # type: typing.Optional[Stream]
    self.v5 = None  # type: typing.Optional[Stream]

  def initialize(self, logs=None, verbosity=None, formatter=None):
    """
    :param list[str] logs:
    :param list[int] verbosity:
    :param list[str] formatter: 'default', 'timed', 'raw' or 'verbose'
    """
    if formatter is None:
      formatter = []
    if verbosity is None:
      verbosity = []
    if logs is None:
      logs = []
    self.initialized = True
    fmt = {
      'default': logging.Formatter('%(message)s'),
      'timed': logging.Formatter('%(asctime)s %(message)s', datefmt='%Y-%m-%d,%H:%M:%S.%MS'),
      'raw': logging.Formatter('%(message)s'),
      'verbose': logging.Formatter('%(levelname)s - %(asctime)s %(message)s', datefmt='%Y-%m-%d,%H:%M:%S.%MS')
    }
    self.v = [logging.getLogger('v' + str(v)) for v in range(6)]
    for l in self.v:
      # Reset handler list, in case we have initialized some earlier (e.g. multiple log.initialize() calls).
      l.handlers = []
    if 'stdout' not in logs:
      logs.append('stdout')
    if len(formatter) == 1:
      # if only one format provided, use it for all logs
      formatter = [formatter[0]] * len(logs)
    for i in range(len(logs)):
      t = logs[i]
      v = 3
      if i < len(verbosity):
        v = verbosity[i]
      elif len(verbosity) == 1:
        v = verbosity[0]
      assert v <= 5, "invalid verbosity: " + str(v)
      f = fmt['default'] if i >= len(formatter) or formatter[i] not in fmt else fmt[formatter[i]]
      if t == 'stdout':
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
      elif t.startswith("|"):  # pipe-format
        proc_cmd = t[1:].strip()
        from subprocess import Popen, PIPE
        proc = Popen(proc_cmd, shell=True, stdin=PIPE)
        handler = logging.StreamHandler(proc.stdin)
        handler.setLevel(logging.DEBUG)
      elif os.path.isdir(os.path.dirname(t)):
        if "$" in t:
          from Util import get_utc_start_time_filename_part
          t = string.Template(t).substitute(date=get_utc_start_time_filename_part())
        self.filename = t
        handler = logging.FileHandler(t)
        handler.setLevel(logging.DEBUG)
      else:
        assert False, "invalid log target %r" % t
      handler.setFormatter(f)
      for j in range(v + 1):
        if handler not in self.v[j].handlers:
          self.v[j].addHandler(handler)
    self.verbose = [True] * 6
    null = logging.FileHandler(os.devnull)
    for i in range(len(self.v)):
      self.v[i].setLevel(logging.DEBUG)
      if not self.v[i].handlers:
        self.verbose[i] = False
        self.v[i].addHandler(null)
    self.error = Stream(self.v[0], logging.CRITICAL)
    self.v0 = Stream(self.v[0], logging.ERROR)
    self.v1 = Stream(self.v[1], logging.INFO)
    self.v2 = Stream(self.v[2], logging.INFO)
    self.v3 = Stream(self.v[3], logging.DEBUG)
    self.v4 = Stream(self.v[4], logging.DEBUG)
    self.v5 = Stream(self.v[5], logging.DEBUG)

  def init_by_config(self, config):
    """
    :param Config.Config config:
    """
    logs = config.list('log', [])
    log_verbosity = config.int_list('log_verbosity', [])
    log_format = config.list('log_format', [])
    if config.is_true("use_horovod"):
      # noinspection PyPackageRequirements,PyUnresolvedReferences
      import horovod.tensorflow as hvd
      from TFUtil import init_horovod
      init_horovod()  # make sure it is initialized
      new_logs = []
      for fn in logs:
        fn_prefix, fn_ext = os.path.splitext(fn)
        fn_ext = ".horovod-%i-%i%s" % (hvd.rank(), hvd.size(), fn_ext)
        new_logs.append(fn_prefix + fn_ext)
      logs = new_logs
    self.initialize(logs=logs, verbosity=log_verbosity, formatter=log_format)


log = Log()

# Some external code (e.g. pyzmq) will initialize the logging system
# via logging.basicConfig(). That will set a handler to the root logger,
# if there is none. This default handler usually prints to stderr.
# Because all our custom loggers are childs of the root logger,
# this will have the effect that everything gets logged twice.
# By adding a dummy handler to the root logger, we will avoid that
# it adds any other default handlers.
logging.getLogger().addHandler(logging.NullHandler())


class StreamThreadLocal(threading.local):
  """
  This will just buffer everything, thread-locally, and not forward it to any stream.
  The idea is that multiple tasks will run in multiple threads and you want to catch all the logging/stdout
  of each to not clutter the output, and also you want to keep it separate for each.
  """

  def __init__(self):
    self.buf = StringIO.StringIO()

  def write(self, msg):
    """
    :param str msg:
    """
    self.buf.write(msg)

  def flush(self):
    """
    Ignored.
    """


class StreamDummy:
  """
  This will just discard any data.
  """

  def write(self, msg):
    """
    Ignored.

    :param str msg:
    """
    pass

  def flush(self):
    """
    Ignored.
    """


@contextlib.contextmanager
def wrap_log_streams(alternative_stream, also_sys_stdout=False, tf_log_verbosity=None):
  """
  :param StreamThreadLocal|StreamDummy alternative_stream:
  :param bool also_sys_stdout: wrap sys.stdout as well
  :param int|str|None tf_log_verbosity: e.g. "WARNING"
  :return: context manager which yields (original info stream v1, alternative_stream)
  """
  v_attrib_keys = ["v%i" % i for i in range(6)] + ["error"]
  # Store original values.
  orig_v_list = log.v
  orig_v_attribs = {key: getattr(log, key) for key in v_attrib_keys}
  orig_stdout = sys.stdout
  log.v = [alternative_stream] * len(orig_v_list)
  for key in v_attrib_keys:
    setattr(log, key, alternative_stream)
  if also_sys_stdout:
    sys.stdout = alternative_stream
  orig_tf_log_verbosity = None
  if tf_log_verbosity is not None:
    import tensorflow as tf
    orig_tf_log_verbosity = tf.logging.get_verbosity()
    tf.logging.set_verbosity(tf_log_verbosity)
  try:
    yield (orig_v_attribs["v1"], alternative_stream)
  finally:
    # Restore original values.
    log.v = orig_v_list
    for key, value in orig_v_attribs.items():
      setattr(log, key, value)
    if also_sys_stdout:
      sys.stdout = orig_stdout
    if tf_log_verbosity is not None:
      import tensorflow as tf
      tf.logging.set_verbosity(orig_tf_log_verbosity)




from __future__ import print_function

import os
import theano
import theano.tensor as T
import numpy
from Log import log
import theano.sandbox.cuda as theano_cuda
from collections import OrderedDict


debug_function_hook = None

def make_fwd_fun(recurrent_transform):
  y_p = recurrent_transform.y_p
  z_re, state_updates = recurrent_transform.step(y_p)
  custom_vars = recurrent_transform.get_sorted_custom_vars()
  state_vars = recurrent_transform.get_sorted_state_vars()

  z_re_shared = recurrent_transform.layer.shared(value=numpy.zeros((1,1),dtype="float32"), name="fwd_fun_z_re_shared")
  updates = [(z_re_shared, z_re)]
  custom_out = []
  state_shared_vars = {v: recurrent_transform.layer.shared(value=numpy.zeros((1,) * v.ndim, dtype="float32"), name=v.name) for v in state_vars}
  for v in state_vars:
    v_upd = state_updates[v]
    updates += [(state_shared_vars[v], v_upd)]
    custom_out += [state_shared_vars[v]]
  fwd_fun = theano.function(inputs=[y_p] + custom_vars + state_vars, outputs=[],
                            updates=updates, on_unused_input="ignore")
  if debug_function_hook:
    fwd_fun = debug_make_theano_function_wrapper(fwd_fun, "att_%i_fwd" % id(recurrent_transform), debug_function_hook, state_shared_vars.values())
  return fwd_fun, z_re_shared, custom_out


def make_bwd_fun(recurrent_transform):
  y_p = recurrent_transform.y_p
  z_re, state_updates = recurrent_transform.step(y_p)
  custom_vars = recurrent_transform.get_sorted_custom_vars()
  state_vars_prev = recurrent_transform.get_sorted_state_vars()

  Dz_re = recurrent_transform.tt.fmatrix("Dz_re")
  state_var_new_grads = {state_updates[v]: v.type("D_" + v.name) for v in state_vars_prev}
  state_var_new_grads_list = [state_var_new_grads[state_updates[k]] for k in state_vars_prev]
  known_grads = {z_re: Dz_re}
  known_grads.update(state_var_new_grads)
  if recurrent_transform.force_gpu:
    # We need the symbolic host representation.
    # See HostFromGpu.grad(). It expects that the output_grads are on the host, i.e. from type T.TensorType.
    # When this is taken out of known_grads, it will fail because they are all CudaNdarrayType.
    # This should anyway be optimized all away and fully taken to the GPU in the final function.
    for k, v in known_grads.items():
      known_grads[k] = theano_cuda.host_from_gpu(v)

  all_wrt = [y_p] + custom_vars + state_vars_prev
  all_grads = T.grad(None, all_wrt, known_grads=OrderedDict(known_grads), disconnected_inputs="ignore")
  assert len(all_grads) == 1 + len(custom_vars) + len(state_vars_prev)
  Dy_p = all_grads[0]
  custom_grads = all_grads[1:len(custom_vars)+1]
  state_var_prev_grads = all_grads[len(custom_vars)+1:]

  out_Dy_p = recurrent_transform.layer.shared(value=numpy.zeros((1,1),dtype="float32"), name="out_Dy_p")
  out_custom_grads = [recurrent_transform.layer.shared(value=numpy.zeros([1] * var.ndim, dtype="float32"), name="out_D_" + var.name) for var in custom_vars]
  out_state_var_prev_grads = [recurrent_transform.layer.shared(value=numpy.zeros([1] * var.ndim, dtype="float32"), name="out_D_" + var.name) for var in state_vars_prev]

  updates = [(out_Dy_p, Dy_p)]
  updates += [(out, out + grad) for out, grad in zip(out_custom_grads, custom_grads)]  # we accumulate the custom input grads
  updates += [(out, grad) for out, grad in zip(out_state_var_prev_grads, state_var_prev_grads)]
  bwd_fun = theano.function(inputs=[y_p] + custom_vars + state_vars_prev + [Dz_re] + state_var_new_grads_list,
                            outputs=[],
                            updates=updates,
                            on_unused_input="ignore")

  # Before we can accumulate the custom input grads, we need to initialize them with 0.
  custom_reset_updates = [(out, T.zeros_like(var)) for out, var in zip(out_custom_grads, custom_vars)]
  custom_reset_fn = theano.function(inputs=custom_vars, outputs=None, updates=custom_reset_updates)

  if debug_function_hook:
    bwd_fun = debug_make_theano_function_wrapper(bwd_fun, "att_%i_bwd" % id(recurrent_transform), debug_function_hook, [])
  return bwd_fun, custom_reset_fn, out_Dy_p, out_custom_grads + out_state_var_prev_grads


def print_wt(op,x):
  print(x.argmax(axis=0))


functions = {}

def setup_parent_functions(fn, recurrent_transform_id):
  import RecurrentTransform
  fn = "_".join(fn.split('_')[:-2])
  if fn in functions: return
  print("loading function", fn, "(pid %i)" % os.getpid(), file=log.v4)
  transform = RecurrentTransform.transforms_by_id[recurrent_transform_id]
  # New instance for the custom op.
  transform = transform.copy_for_custom()
  assert isinstance(transform, RecurrentTransform.RecurrentTransformBase)
  _setup_func(fn, transform)
  functions[fn] = transform

def _setup_func(fn, recurrent_transform):
  fwd_names = ["_fun_fwd", "_fun_fwd_res0", "_fun_fwd_res1"]
  bwd_names = ["_fun_bwd", "_fun_reset", "_fun_bwd_res0", "_fun_bwd_res1"]
  vs_fwd = make_fwd_fun(recurrent_transform)
  vs_bwd = make_bwd_fun(recurrent_transform)
  assert len(vs_fwd) == len(fwd_names)
  assert len(vs_bwd) == len(bwd_names)
  for v, postfix in zip(vs_fwd + vs_bwd, fwd_names + bwd_names):
    globals()[fn + postfix] = v

def debug_make_theano_function_wrapper(f, name, hook, other_values):
  def to_str(v):
    if isinstance(v, (list, tuple)):
      return "[%s]" % ", ".join([to_str(v0) for v0 in v])
    if isinstance(v, theano.compile.SharedVariable):
      return "%s = %s" % (str(v), to_str(v.get_value(borrow=True, return_internal_type=True)))
    v = numpy.asarray(v)
    if len(v.shape) >= 2:
      return "\n" + str(v)
    return str(v)
  def theano_func_wrapped(*args):
    res = f(*args)
    print("called", name, "args:", to_str(args), "res:", to_str(res), "other:", to_str(other_values))
    if hook and hook is not True:
      hook(f=f, name=name, args=args, res=res)
    return res
  return theano_func_wrapped



import os
from FunctionLoader import make_funloader_code
import theano
import theano.gradient
import theano.tensor as T
import theano.printing
import theano.gof
from theano.gof.opt import OpSub
from theano.compile import optdb
from theano.sandbox.cuda.basic_ops import (as_cuda_ndarray_variable,
                                           gpu_contiguous)


class LSTMRecOpGrad(theano.sandbox.cuda.GpuOp):
  __props__ = ("inplace", "fun_name", "recurrent_transform")

  def __init__(self, fun_name, inplace, recurrent_transform):
    """
    :type recurrent_transform: RecurrentTransform.RecurrentTransformBase
    """
    super(LSTMRecOpGrad, self).__init__(self)
    self.inplace = inplace
    self.fun_name = fun_name
    self.recurrent_transform = recurrent_transform
    if inplace:
     # http://deeplearning.net/software/theano/extending/inplace.html
     # https://github.com/Theano/Theano/issues/3506
     # It's strange that we must mark which output operates on which input -
     # I would expect that it must only know which inputs are destroyed.
     # Anyway:
     # All outputs operate inplace on inputs 1 and 6 (which are H and DY)
     # but when the input is marked multiple times, we get an error.
     # This is also strange, and probably a bug in Theano.
     # So we could mark that output 0 destroys inputs 1 and 6.
     # That also doesn't work, it will not apply the inplace-optimization anymore.
     # So, we do it in some other way. From how I understand the Theano code,
     # the output index is ignored, so we can use any.
     # Anyway Theano knows what inputs will be destroyed, so it should be OK.
     destroy_input_list = [1, 6]
     self.destroy_map = {i: [i] for i in destroy_input_list}  # hack, see above

  def _get_num_custom_vars(self):
    return len(self.recurrent_transform.custom_vars)

  def _get_num_state_vars(self):
    return len(self.recurrent_transform.state_vars)

  def make_node(self, Y, H, c, y0, i, Dd, DY, *args):
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    y0 = gpu_contiguous(as_cuda_ndarray_variable(y0))
    i = gpu_contiguous(as_cuda_ndarray_variable(T.cast(i,'float32')))
    Dd = gpu_contiguous(as_cuda_ndarray_variable(Dd))
    DY = gpu_contiguous(as_cuda_ndarray_variable(DY))
    args = [gpu_contiguous(as_cuda_ndarray_variable(x)) for x in args]
    # args = custom_inputs + state_vars_seqs
    assert len(args) == self._get_num_custom_vars() + self._get_num_state_vars()
    assert DY.dtype == 'float32'
    assert Y.dtype == 'float32'
    assert H.dtype == 'float32'
    assert c.dtype == 'float32'
    assert y0.dtype == "float32"
    for x in args:
      assert x.dtype == "float32"
    assert DY.ndim == 3
    assert Y.ndim == 3
    assert H.ndim == 3
    assert c.ndim == 2
    assert y0.ndim == 2
    assert i.ndim == 2

    custom_input_grads = [var.type() for var in args[:self._get_num_custom_vars()]]
    CudaNdarrayType = theano.sandbox.cuda.CudaNdarrayType
    # One ndim less because initial state var grads vs whole seq state vars.
    initial_state_var_grads = [CudaNdarrayType(dtype="float32", broadcastable=(False,) * (var.ndim - 1))()
                               for var in args[self._get_num_custom_vars():]]
    return theano.Apply(self, [Y, H, c, y0, i, Dd, DY] + args,
                        # DZ, Dc, Dy0, custom input grads, initial state var grads
                        [H.type(), c.type(), y0.type()] + custom_input_grads + initial_state_var_grads)

  def c_support_code(self):
    crnn_path = os.path.dirname(__file__)
    fun_prefix = "%s_%i" % (self.fun_name, id(self.recurrent_transform))
    funloader = make_funloader_code(self.recurrent_transform, fun_prefix + "_fun_bwd", fun_prefix + "_fun_reset")
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return funloader + f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    (Y, H, c, y0, i, Dd, DY), remaining_inputs = input_names[:7], input_names[7:]
    assert len(remaining_inputs) == self._get_num_custom_vars() + self._get_num_state_vars()
    custom_inputs = remaining_inputs[:self._get_num_custom_vars()]
    seq_state_var_names = remaining_inputs[self._get_num_custom_vars():]
    custom_inputs_str = ",".join(custom_inputs)
    seq_state_var_names_str = ", ".join(seq_state_var_names)
    (DZ, Dc, Dy0), remaining_outputs = output_names[:3], output_names[3:]
    assert len(remaining_outputs) == self._get_num_custom_vars() + self._get_num_state_vars()
    custom_output_names = remaining_outputs[:self._get_num_custom_vars()]
    initial_state_var_grad_names = remaining_outputs[self._get_num_custom_vars():]
    custom_outputs_str = ", ".join(["&" + grad for grad in custom_output_names])
    initial_state_var_grad_names_str = ", ".join(["&" + grad for grad in initial_state_var_grad_names])
    bwd_fun = "%s_%i_fun_bwd" % (self.fun_name, id(self.recurrent_transform))
    fail = sub['fail']
    inplace = "true" if self.inplace else "false"
    return """
    //std::cout << "LSTMRecOpGrad called" << std::endl;
    if(!%(inplace)s)
    {
      //std::cout << "warning, inplace optimization failed, not working inplace" << std::endl;
    }

    if(%(DZ)s || %(Dc)s || %(Dy0)s)
    {
      printf("output storage already exists\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(DZ)s);
      Py_XDECREF(%(Dc)s);
      Py_XDECREF(%(Dy0)s);
    }

    #define ARRAY_LEN(x) (sizeof(x) / sizeof(x[0]))
    CudaNdarray* custom_inputs[] = {%(custom_inputs_str)s}; // input
    %(bwd_fun)s.reset_shared(custom_inputs, ARRAY_LEN(custom_inputs)); // init the custom grads with zero

    CudaNdarray* seq_state_vars[] = {%(seq_state_var_names_str)s}; // input
    CudaNdarray** state_var_grads[] = {%(initial_state_var_grad_names_str)s}; // output
    for(int i = 0; i < ARRAY_LEN(state_var_grads); ++i) {
      Py_XDECREF(*state_var_grads[i]); // in case of earlier output storage
      // dims like seq_state_vars[i] without time, which is the first dim
      int ndim = CudaNdarray_NDIM(seq_state_vars[i]) - 1;
      const int* dims = CudaNdarray_HOST_DIMS(seq_state_vars[i]) + 1;
      *state_var_grads[i] = (CudaNdarray*) CudaNdarray_ZEROS(ndim, (int*) dims);
      assert(*state_var_grads[i]);
    }

    CudaNdarray * epsilon = 0;
    CudaNdarray * delta = 0;
    if(%(inplace)s)
    {
      epsilon = %(DY)s;
      delta = %(H)s;
      Py_XINCREF(delta);
    }
    else
    {
      epsilon = (CudaNdarray *) CudaNdarray_Copy(%(DY)s);
      delta = (CudaNdarray *) CudaNdarray_Copy(%(H)s);
    }

    const int * H_dim = CudaNdarray_HOST_DIMS(%(H)s);
    const int * Y_dim = CudaNdarray_HOST_DIMS(%(Y)s);

    int y = 0;
    for(int x = H_dim[0]-1; x >= 0; --x)
    {
      //add recurrent
      bool rightBorder = (x == H_dim[0]-1);
      bool leftBorder = (x == 0);

      //call custom function here
      if(!rightBorder)
      {
        CudaNdarray * y_p = 0;
        //x-1?
        PyObject * y_p_obj = PyObject_CallMethod((PyObject*) %(Y)s, "__getitem__", "(i)", x);
        assert(y_p_obj);
        y_p = (CudaNdarray*) y_p_obj;

        PyObject * delta_x_obj = PyObject_CallMethod((PyObject*) delta, "__getitem__", "(i)", x+1);
        assert(delta_x_obj);
        CudaNdarray * delta_x = (CudaNdarray*) delta_x_obj;

        CudaNdarray* state_vars_prev[ARRAY_LEN(seq_state_vars)];
        for(int i = 0; i < ARRAY_LEN(seq_state_vars); ++i) {
          state_vars_prev[i] = (CudaNdarray*) PyObject_CallMethod((PyObject*) seq_state_vars[i], "__getitem__", "(i)", x+1);
          assert(state_vars_prev[i]);
        }

        // bwd_fun args: y_p, custom inputs, state vars prev, Dz_re, state var new grads
        CudaNdarray* bwd_fun_inputs[2 + ARRAY_LEN(custom_inputs) + 2 * ARRAY_LEN(seq_state_vars)];
        {
          int idx = 0;
          bwd_fun_inputs[idx++] = y_p;
          for(int i = 0; i < ARRAY_LEN(custom_inputs); ++i)
            bwd_fun_inputs[idx++] = custom_inputs[i];
          for(int i = 0; i < ARRAY_LEN(state_vars_prev); ++i)
            bwd_fun_inputs[idx++] = state_vars_prev[i];
          bwd_fun_inputs[idx++] = delta_x;
          for(int i = 0; i < ARRAY_LEN(state_var_grads); ++i)
            bwd_fun_inputs[idx++] = *state_var_grads[i];
          assert(idx == ARRAY_LEN(bwd_fun_inputs));
        }
        std::vector<CudaNdarray*> res_vec = %(bwd_fun)s.call(bwd_fun_inputs, ARRAY_LEN(bwd_fun_inputs));
        // result shared vars: Dy_p, custom input grads, state var grads
        assert(res_vec.size() == 1 + ARRAY_LEN(custom_inputs) + ARRAY_LEN(seq_state_vars));
        Py_XDECREF(delta_x);
        CudaNdarray * Dy_p = (CudaNdarray*) res_vec[0];

        //copy to epsilon
        float * epsilon_x_data = data_ptr(epsilon, y, x);
        do_add(epsilon_x_data, CudaNdarray_DEV_DATA(Dy_p), CudaNdarray_SIZE(Dy_p));

        // custom input grads will automatically be accumulated. see CustomLSTMFunctions.
        // copy state var grads
        {
          int idx = 1 + ARRAY_LEN(custom_inputs);
          for(int i = 0; i < ARRAY_LEN(seq_state_vars); ++i) {
            CudaNdarray* dst = *state_var_grads[i];
            CudaNdarray* src = res_vec[idx++];
            assert(CudaNdarray_SIZE(dst) == CudaNdarray_SIZE(src));
            cudaMemcpy(
              CudaNdarray_DEV_DATA(dst), CudaNdarray_DEV_DATA(src),
              CudaNdarray_SIZE(src) * sizeof(real), cudaMemcpyDeviceToDevice);
          }
          assert(res_vec.size() == idx);
        }

        for(int i = 0; i < res_vec.size(); ++i)
          Py_XDECREF(res_vec[i]);
        for(int i = 0; i < ARRAY_LEN(state_vars_prev); ++i)
          Py_XDECREF(state_vars_prev[i]);
        Py_XDECREF(y_p);
      }

      do_lstm_bwd(delta, epsilon, %(Y)s, %(Dd)s, %(c)s, y, x, rightBorder, %(i)s);
    }

    %(DZ)s = delta;

    %(Dc)s = CudaNdarray_uninitialized_like(%(c)s);
    HANDLE_ERROR(cudaMemcpy(CudaNdarray_DEV_DATA(%(Dc)s), CudaNdarray_DEV_DATA(epsilon),
      Y_dim[1]*Y_dim[2]*sizeof(float), cudaMemcpyDeviceToDevice));

    %(Dy0)s = CudaNdarray_zeros_like(%(y0)s);
    //add custom function
    //TODO: move to function
    PyObject * delta_x_obj = PyObject_CallMethod((PyObject*) delta, "__getitem__", "(i)", 0);
    assert(delta_x_obj);
    CudaNdarray * delta_x = (CudaNdarray*) delta_x_obj;

    CudaNdarray* state_vars_prev[ARRAY_LEN(seq_state_vars)];
    for(int i = 0; i < ARRAY_LEN(seq_state_vars); ++i) {
      // left border
      state_vars_prev[i] = (CudaNdarray*) PyObject_CallMethod((PyObject*) seq_state_vars[i], "__getitem__", "(i)", 0);
      assert(state_vars_prev[i]);
    }

    // bwd_fun args: y_p, custom inputs, state vars prev, Dz_re, state var new grads
    CudaNdarray* bwd_fun_inputs[2 + ARRAY_LEN(custom_inputs) + 2 * ARRAY_LEN(seq_state_vars)];
    {
      int idx = 0;
      bwd_fun_inputs[idx++] = %(y0)s;
      for(int i = 0; i < ARRAY_LEN(custom_inputs); ++i)
        bwd_fun_inputs[idx++] = custom_inputs[i];
      for(int i = 0; i < ARRAY_LEN(state_vars_prev); ++i)
        bwd_fun_inputs[idx++] = state_vars_prev[i];
      bwd_fun_inputs[idx++] = delta_x;
      for(int i = 0; i < ARRAY_LEN(state_var_grads); ++i)
        bwd_fun_inputs[idx++] = *state_var_grads[i];
      assert(idx == ARRAY_LEN(bwd_fun_inputs));
    }
    std::vector<CudaNdarray*> res_vec = %(bwd_fun)s.call(bwd_fun_inputs, ARRAY_LEN(bwd_fun_inputs));
    // result shared vars: Dy_p, custom input grads, state var grads
    assert(res_vec.size() == 1 + ARRAY_LEN(custom_inputs) + ARRAY_LEN(seq_state_vars));
    Py_XDECREF(delta_x);
    {
      int idx = 0;
      CudaNdarray * Dy_p = res_vec[idx++];
      //copy to Dy0
      do_add(CudaNdarray_DEV_DATA(%(Dy0)s), CudaNdarray_DEV_DATA(Dy_p), CudaNdarray_SIZE(Dy_p));


      //custom grads
      CudaNdarray** custom_grads[] = {%(custom_outputs_str)s}; // output
      for(int i = 0; i < ARRAY_LEN(custom_grads); ++i) {
        *custom_grads[i] = (CudaNdarray*) CudaNdarray_Copy(res_vec[idx++]);
        assert(*custom_grads[i]);
      }

      // copy state var grads
      for(int i = 0; i < ARRAY_LEN(seq_state_vars); ++i) {
        CudaNdarray* dst = *state_var_grads[i];
        CudaNdarray* src = res_vec[idx++];
        assert(CudaNdarray_SIZE(dst) == CudaNdarray_SIZE(src));
        cudaMemcpy(
          CudaNdarray_DEV_DATA(dst), CudaNdarray_DEV_DATA(src),
          CudaNdarray_SIZE(src) * sizeof(real), cudaMemcpyDeviceToDevice);
      }
      assert(res_vec.size() == idx);
    }

    for(int i = 0; i < res_vec.size(); ++i)
      Py_XDECREF(res_vec[i]);
    for(int i = 0; i < ARRAY_LEN(state_vars_prev); ++i)
      Py_XDECREF(state_vars_prev[i]);
    if(!%(inplace)s)
      Py_XDECREF(epsilon);

    #undef ARRAY_LEN
    """ % locals()

#------------------------

class LSTMRecOp(theano.sandbox.cuda.GpuOp):
  __props__ = ("inplace", "fun_name", "recurrent_transform")

  def __init__(self, fun_name, inplace, recurrent_transform):
    """
    :type recurrent_transform: RecurrentTransform.RecurrentTransformBase
    """
    super(LSTMRecOp, self).__init__(self)
    self.inplace = inplace
    self.fun_name = fun_name
    self.recurrent_transform = recurrent_transform
    if inplace:
      #all outputs operate inplace on input 0 (which is Z)
      #but when the input is marked multiple times, we get an error
      #so we only mark that output 0 destroys input 0
      #anyway theano knows that input 0 will be destroyed, so it should be OK
      self.destroy_map = {0: [0]}

  def _get_num_custom_vars(self):
    return len(self.recurrent_transform.custom_vars)

  def _get_num_state_vars(self):
    return len(self.recurrent_transform.state_vars)

  def _seq_var_for_initial_state_var(self, v):
    type_class = v.type.__class__
    # One ndim more for time.
    seq_var_type = type_class(dtype="float32", broadcastable=(False,) * (v.ndim + 1))
    return seq_var_type()

  def make_node(self, Z, c, y0, i, *args):
    """
    :param Z: {input,output,forget} gate + cell state. 3d (time,batch,dim*4)
    :param c: initial cell state. 2d (batch,dim)
    :param y0: output of t = -1 (for recursion at t = 0). 2d (batch,dim)
    :param i: index. 2d (time,batch) -> 0 or 1
    :param args: custom_inputs + initial_state_vars: other inputs for the custom function
    """
    from Util import have_gpu
    assert have_gpu()

    assert len(args) == self._get_num_custom_vars() + self._get_num_state_vars(), self.recurrent_transform
    custom_inputs = args[:self._get_num_custom_vars()]
    initial_state_vars = args[self._get_num_custom_vars():]

    custom_inputs = [gpu_contiguous(as_cuda_ndarray_variable(x)) for x in custom_inputs]
    initial_state_vars = [gpu_contiguous(as_cuda_ndarray_variable(x)) for x in initial_state_vars]
    Z = gpu_contiguous(as_cuda_ndarray_variable(Z))
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    y0 = gpu_contiguous(as_cuda_ndarray_variable(y0))
    i = gpu_contiguous(as_cuda_ndarray_variable(T.cast(i,'float32')))
    assert Z.dtype == "float32"
    assert c.dtype == "float32"
    assert y0.dtype == "float32"
    for x in custom_inputs:
      assert x.dtype == "float32"
    for x in initial_state_vars:
      assert x.dtype == "float32"
    assert Z.ndim == 3
    assert c.ndim == 2
    assert y0.ndim == 2
    assert i.ndim == 2

    seq_state_vars = [self._seq_var_for_initial_state_var(x) for x in initial_state_vars]
    return theano.Apply(self,
                        [Z, c, y0, i] + custom_inputs + initial_state_vars,
                        # results: (output) Y, (gates and cell state) H, (final cell state) d, state vars sequences
                        [Z.type(), Z.type(), c.type()] + seq_state_vars)

  def c_support_code(self):
    fun_prefix = "%s_%i" % (self.fun_name, id(self.recurrent_transform))
    funloader = make_funloader_code(self.recurrent_transform, fun_prefix + "_fun_fwd")
    crnn_path = os.path.dirname(__file__)
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return funloader + f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    # Y: all the outputs. 3d (time,batch,dim)
    # Z/H: {input,output,forget} gate + cell state. 3d (time,batch,dim*4)
    # d: last state (= Y[T-1]). 2d (batch,dim)
    Z, c, y0, i = input_names[:4]
    custom_inputs = input_names[4:]
    assert len(custom_inputs) == self._get_num_custom_vars() + self._get_num_state_vars()
    custom_inputs, initial_state_vars = custom_inputs[:self._get_num_custom_vars()], custom_inputs[self._get_num_custom_vars():]
    custom_inputs_str = ", ".join(custom_inputs)
    initial_state_vars_str = ", ".join(initial_state_vars)
    Y, H, d = output_names[:3]
    state_vars_seqs = output_names[3:]
    assert len(state_vars_seqs) == self._get_num_state_vars()
    state_vars_seqs_str_comma = "".join([", %s[x]" % x for x in state_vars_seqs])
    state_vars_seqs_ptr_str = ", ".join(["&" + x for x in state_vars_seqs])
    fwd_fun = "%s_%i_fun_fwd" % (self.fun_name, id(self.recurrent_transform))
    inplace = "true" if self.inplace else "false"
    fail = sub['fail']
    # see https://github.com/Theano/Theano/blob/master/theano/sandbox/cuda/cuda_ndarray.cuh for some doc
    return """
    if(%(Y)s || %(H)s || %(d)s)
    {
      printf("Y or H or d already exist\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(Y)s);
      Py_XDECREF(%(H)s);
      Py_XDECREF(%(d)s);
    }

    // outputs
    const int * Z_dim = CudaNdarray_HOST_DIMS(%(Z)s);
    const int dims_Y[] = {Z_dim[0], Z_dim[1], Z_dim[2] / 4};
    const int dims_H[] = {Z_dim[0], Z_dim[1], Z_dim[2]};
    const int dims_d[] = {Z_dim[1], Z_dim[2] / 4};
    int size_d = Z_dim[1] * Z_dim[2] / 4;

    %(Y)s = (CudaNdarray*) CudaNdarray_NewDims(3, dims_Y);
    %(d)s = (CudaNdarray*) CudaNdarray_NewDims(2, dims_d);
    if(%(inplace)s)
    {
      %(H)s = %(Z)s;
      Py_INCREF(%(Z)s);
    }
    else
    {
      %(H)s = (CudaNdarray*) CudaNdarray_NewDims(3,dims_H);
      cudaMemcpy(CudaNdarray_DEV_DATA(%(H)s), CudaNdarray_DEV_DATA(%(Z)s),
      dims_H[0]*dims_H[1]*dims_H[2]*sizeof(float), cudaMemcpyDeviceToDevice);
    }

    CudaNdarray* custom_inputs[] = {%(custom_inputs_str)s};

    // custom state vars seqs outputs
    #define ARRAY_LEN(x) (sizeof(x) / sizeof(x[0]))
    CudaNdarray* initial_state_vars[] = {%(initial_state_vars_str)s};
    CudaNdarray** state_vars_seqs_ptr[] = {%(state_vars_seqs_ptr_str)s};
    assert(ARRAY_LEN(initial_state_vars) == ARRAY_LEN(state_vars_seqs_ptr));
    for(int i = 0; i < ARRAY_LEN(initial_state_vars); ++i) {
      const int initial_ndim = CudaNdarray_NDIM(initial_state_vars[i]);
      int ndim = initial_ndim + 1; // add time-dim
      const int* initial_dims = CudaNdarray_HOST_DIMS(initial_state_vars[i]);
      int dims[] = {Z_dim[0], 0, 0, 0};
      assert(ARRAY_LEN(dims) >= ndim);
      for(int d = 0; d < initial_ndim; ++d)
        dims[d + 1] = initial_dims[d];
      *state_vars_seqs_ptr[i] = (CudaNdarray*) CudaNdarray_NewDims(ndim, dims);
      // copy initial over
      cudaMemcpy(
        CudaNdarray_DEV_DATA(*state_vars_seqs_ptr[i]),
        CudaNdarray_DEV_DATA(initial_state_vars[i]),
        CudaNdarray_SIZE(initial_state_vars[i]) * sizeof(real), cudaMemcpyDeviceToDevice);
    }

    int y = 0;
    for(int x = 0; x < Z_dim[0]; ++x)
    {
      bool leftBorder = (x == 0);
      bool rightBorder = (x == Z_dim[0] - 1);
      // call custom function here

      CudaNdarray* state_vars[ARRAY_LEN(state_vars_seqs_ptr)];
      for(int i = 0; i < ARRAY_LEN(state_vars_seqs_ptr); ++i) {
        state_vars[i] = (CudaNdarray*) PyObject_CallMethod((PyObject*) *state_vars_seqs_ptr[i], "__getitem__", "(i)", x);
        assert(state_vars[i]);
      }

      CudaNdarray * y_p = 0;
      if(leftBorder)
      {
        y_p = %(y0)s;
      }
      else
      {
        PyObject * y_p_obj = PyObject_CallMethod((PyObject*) %(Y)s, "__getitem__", "(i)", x-1);
        assert(y_p_obj);
        y_p = (CudaNdarray*) y_p_obj;
      }
      //std::cerr << "t=" << x << std::endl;
      // fwd fun args: y_p, custom inputs, state vars
      CudaNdarray* fun_args[1 + ARRAY_LEN(custom_inputs) + ARRAY_LEN(state_vars)];
      {
        int idx = 0;
        fun_args[idx++] = y_p;
        for(int i = 0; i < ARRAY_LEN(custom_inputs); ++i)
          fun_args[idx++] = custom_inputs[i];
        for(int i = 0; i < ARRAY_LEN(state_vars); ++i)
          fun_args[idx++] = state_vars[i];
        assert(idx == ARRAY_LEN(fun_args));
      }
      std::vector<CudaNdarray*> res_vec = %(fwd_fun)s.call(fun_args, ARRAY_LEN(fun_args));
      assert(res_vec.size() == 1 + ARRAY_LEN(initial_state_vars));

      // add to H
      {
        CudaNdarray * res = res_vec[0];
        float * H_y_x_data = data_ptr(%(H)s, y, x);
        do_add(H_y_x_data, CudaNdarray_DEV_DATA(res), CudaNdarray_SIZE(res));
      }

      if(!rightBorder) {
        // set new state vars
        for(int i = 0; i < ARRAY_LEN(initial_state_vars); ++i) {
          CudaNdarray* src = res_vec[i + 1];
          float* src_ptr = CudaNdarray_DEV_DATA(src);
          CudaNdarray* dst = *state_vars_seqs_ptr[i];
          float* dst_ptr = CudaNdarray_DEV_DATA(dst) + CudaNdarray_HOST_STRIDES(dst)[0] * (x + 1);
          assert(CudaNdarray_HOST_STRIDES(dst)[0] == CudaNdarray_SIZE(src));
          cudaMemcpy(dst_ptr, src_ptr, CudaNdarray_SIZE(src) * sizeof(real), cudaMemcpyDeviceToDevice);
        }
      }

      for(int i = 0; i < res_vec.size(); ++i)
        Py_XDECREF(res_vec[i]);
      if(!leftBorder)
        Py_XDECREF(y_p);
      for(int i = 0; i < ARRAY_LEN(state_vars); ++i)
        Py_XDECREF(state_vars[i]);


      float * d_ptr = rightBorder ? CudaNdarray_DEV_DATA(%(d)s) : 0;
      do_lstm(%(H)s, %(Y)s, %(c)s, d_ptr, y, x, %(i)s);
    }
    #undef ARRAY_LEN
    """ % locals()

  def grad(self, inputs, output_grads):
    (Z, c, y0, index), input_rest = inputs[:4], inputs[4:]
    assert len(input_rest) == self._get_num_custom_vars() + self._get_num_state_vars()
    custom_inputs = input_rest[:self._get_num_custom_vars()]
    initial_state_vars = input_rest[self._get_num_custom_vars():]
    (DY, DH, Dd), seq_state_var_grads = output_grads[:3], output_grads[3:]
    assert len(seq_state_var_grads) == self._get_num_state_vars()

    Z_raw = Z.owner.inputs[0].owner.inputs[0]
    c_raw = c.owner.inputs[0].owner.inputs[0]
    y0_raw = y0.owner.inputs[0].owner.inputs[0]
    i_raw = index.owner.inputs[0].owner.inputs[0]
    custom_inputs_raw = [x.owner.inputs[0] for x in custom_inputs]
    #we have to make sure that this in only computed once!
    #for this we have to extract the raw variables before conversion to continuous gpu array
    #so that theano can merge the nodes
    all_out = self(*([Z_raw, c_raw, y0_raw, i_raw] + custom_inputs + initial_state_vars))
    (Y, H, d), seq_state_vars = all_out[:3], all_out[3:]

    assert isinstance(DH.type, theano.gradient.DisconnectedType)  # DH is ignored.
    if isinstance(DY.type, theano.gradient.DisconnectedType):
      DY = T.zeros_like(Z)
    if isinstance(Dd.type, theano.gradient.DisconnectedType):
      Dd = T.zeros_like(c)
    for i in range(len(seq_state_var_grads)):
      if isinstance(seq_state_var_grads[i].type, theano.gradient.DisconnectedType):
        # First dim for time. One more for -1 element.
        shape = [Z.shape[0] + 1] + [initial_state_vars[i].shape[d] for d in range(initial_state_vars[i].ndim)]
        seq_state_var_grads[i] = T.zeros(shape, dtype="float32")

    grad_op = grad_ops[(self.fun_name, id(self.recurrent_transform))]
    all_grads = grad_op(*([Y, H, c, y0, index, Dd, DY] + custom_inputs + seq_state_var_grads))
    (DZ, Dc, Dy0), remaining_grads = all_grads[:3], all_grads[3:]
    # remaining grads = custom_inputs grads + initial state var grads
    assert len(remaining_grads) == self._get_num_custom_vars() + self._get_num_state_vars()
    custom_input_grads = remaining_grads[:self._get_num_custom_vars()]
    initial_state_var_grads = remaining_grads[self._get_num_custom_vars():]
    Di = theano.gradient.grad_undefined(self, 3, inputs[3], 'cannot diff w.r.t. index')

    return [DZ, Dc, Dy0, Di] + custom_input_grads + initial_state_var_grads


function_ops = {}; ":type: dict[(str,int),LSTMRecOp]"
grad_ops = {}; ":type: dict[(str,int),LSTMRecOpGrad]"

def register_func(recurrent_transform):
  """
  :type recurrent_transform: RecurrentTransform.RecurrentTransformBase
  """
  fn = recurrent_transform.name
  key = (fn, id(recurrent_transform))
  if key in function_ops:
    return function_ops[key]

  # register op
  no_inpl = LSTMRecOp(fun_name=fn, inplace=False, recurrent_transform=recurrent_transform)
  inpl = LSTMRecOp(fun_name=fn, inplace=True, recurrent_transform=recurrent_transform)
  function_ops[key] = no_inpl

  # hack to avoid being called twice
  attr = 'LSTMRecOpInplaceOpt_%s_%i' % (fn, id(recurrent_transform))
  if not hasattr(optdb, attr):
    opt = OpSub(no_inpl, inpl)
    optdb.register(attr, theano.gof.TopoOptimizer(opt),
                   50.0, 'fast_run', 'inplace', 'gpuarray')
    setattr(optdb, attr, True)

  # the same for grad
  no_inpl = LSTMRecOpGrad(fun_name=fn, inplace=False, recurrent_transform=recurrent_transform)
  inpl = LSTMRecOpGrad(fun_name=fn, inplace=True, recurrent_transform=recurrent_transform)
  grad_ops[key] = no_inpl

  # hack to avoid being called twice
  attr = 'LSTMRecOpGradInplaceOpt_%s_%i' % (fn, id(recurrent_transform))
  if not hasattr(optdb, attr):
    opt = OpSub(no_inpl, inpl)
    optdb.register(attr, theano.gof.TopoOptimizer(opt),
                   50.0, 'fast_run', 'inplace', 'gpuarray')
    setattr(optdb, attr, True)

  return function_ops[key]



#!/usr/bin/env python

from __future__ import print_function

import os
import sys

my_dir = os.path.dirname(os.path.abspath(__file__))
returnn_dir = os.path.dirname(my_dir)
sys.path.append(returnn_dir)

import argparse
import rnn
import json
from Log import log
from Pretrain import pretrain_from_config
from Network import LayerNetwork


def init(configFilename, commandLineOptions):
  rnn.init_better_exchook()
  rnn.init_config(configFilename, commandLineOptions)
  global config
  config = rnn.config
  config.set("log", [])
  rnn.init_log()
  print("CRNN dump-dataset starting up.", file=log.v3)
  rnn.init_config_json_network()


def main(argv):
  argparser = argparse.ArgumentParser(description='Dump network as JSON.')
  argparser.add_argument('crnn_config_file')
  argparser.add_argument('--epoch', default=1, type=int)
  argparser.add_argument('--out', default="/dev/stdout")
  args = argparser.parse_args(argv[1:])
  init(configFilename=args.crnn_config_file, commandLineOptions=[])

  pretrain = pretrain_from_config(config)
  if pretrain:
    network = pretrain.get_network_for_epoch(args.epoch)
  else:
    network = LayerNetwork.from_config_topology(config)

  json_data = network.to_json_content()
  f = open(args.out, 'w')
  print(json.dumps(json_data, indent=2, sort_keys=True), file=f)
  f.close()

  rnn.finalize()


if __name__ == '__main__':
  main(sys.argv)



#!/usr/bin/env python3


from __future__ import print_function

import os
import sys
import time
import tensorflow as tf
import numpy

my_dir = os.path.dirname(os.path.abspath(__file__))
returnn_dir = os.path.dirname(my_dir)
sys.path.insert(0, returnn_dir)

import rnn
from Log import log
import argparse
from Util import Stats, hms
from Dataset import Dataset, init_dataset
import Util
import TFUtil


def init(config_filename, log_verbosity):
  """
  :param str config_filename: filename to config-file
  :param int log_verbosity:
  """
  rnn.init_better_exchook()
  rnn.init_thread_join_hack()
  if config_filename:
    print("Using config file %r." % config_filename)
    assert os.path.exists(config_filename)
  rnn.init_config(config_filename=config_filename, command_line_options=[])
  global config
  config = rnn.config
  config.set("log", None)
  config.set("log_verbosity", log_verbosity)
  config.set("use_tensorflow", True)
  rnn.init_log()
  print("Returnn compile-native-op starting up.", file=log.v1)
  rnn.returnn_greeting()
  rnn.init_backend_engine()
  assert Util.BackendEngine.is_tensorflow_selected(), "this is only for TensorFlow"
  rnn.init_faulthandler()
  rnn.init_config_json_network()
  if 'network' in config.typed_dict:
    print("Loading network")
    from TFNetwork import TFNetwork
    network = TFNetwork(
      name="root",
      config=config,
      rnd_seed=1,
      train_flag=False,
      eval_flag=True,
      search_flag=False)
    network.construct_from_dict(config.typed_dict["network"])


def main(argv):
  from TFUtil import CudaEnv, NativeCodeCompiler
  CudaEnv.verbose_find_cuda = True
  NativeCodeCompiler.CollectedCompilers = []

  argparser = argparse.ArgumentParser(description='Compile some op')
  argparser.add_argument('--config', help="filename to config-file")
  argparser.add_argument('--native_op', help="op name. e.g. 'LstmGenericBase'")
  argparser.add_argument('--blas_lib', default=None,
                         help="specify which blas lib to use (path to .so or file name to search for)")
  argparser.add_argument('--search_for_numpy_blas', dest='search_for_numpy_blas', action='store_true',
                         help="search for blas inside numpys .libs folder")
  argparser.add_argument('--no_search_for_numpy_blas', dest='search_for_numpy_blas', action='store_false',
                         help="do not search for blas inside numpys .libs folder")
  argparser.add_argument("--verbosity", default=4, type=int, help="5 for all seqs (default: 4)")
  argparser.add_argument("--output_file", help='if given, will write the list of libs to this file')
  args = argparser.parse_args(argv[1:])
  init(config_filename=args.config, log_verbosity=args.verbosity)

  import NativeOp
  from TFNativeOp import make_op, OpMaker
  if args.native_op:
    print("Loading native op %r" % args.native_op)
    make_op(getattr(NativeOp, args.native_op), compiler_opts={"verbose": True},
            search_for_numpy_blas=args.search_for_numpy_blas, blas_lib=args.blas_lib)

  libs = []
  if OpMaker.with_cuda and OpMaker.tf_blas_gemm_workaround:
    print('CUDA BLAS lib:', OpMaker.cuda_blas_gemm_so_filename())
    libs.append(OpMaker.cuda_blas_gemm_so_filename())
  elif OpMaker.with_cuda is False:
    print('No CUDA.')

  for compiler in NativeCodeCompiler.CollectedCompilers:
    assert isinstance(compiler, NativeCodeCompiler)
    print(compiler)
    libs.append(compiler._so_filename)

  if libs:
    print("libs:")
    for fn in libs:
      print(fn)
  else:
    print("no libs compiled. use --native_op or --config")

  if args.output_file:
    with open(args.output_file, "w") as f:
      for fn in libs:
        f.write(fn + '\n')
    print("Wrote lib list to file:", args.output_file)


if __name__ == '__main__':
  main(sys.argv)




"""
Contains utility functions to construct a batch.
This is used both by Theano and TF.
"""

from __future__ import print_function

import numpy
from EngineBatch import Batch
from Log import log


def assign_dev_data(device, dataset, batches, load_seqs=True):
  """
  :type device: Device.Device
  :type dataset: Dataset.Dataset
  :type batches: list[EngineBatch.Batch]
  :param bool load_seqs:
  :returns successful and how much batch idx to advance.
  :rtype: (bool,int)
  """
  from Dataset import shapes_for_batches
  shapes = shapes_for_batches(batches, data_keys=device.used_data_keys, dataset=dataset)
  if shapes is None:
    return False, len(batches)
  device.alloc_data(shapes=shapes, max_ctc_length=dataset.get_max_ctc_length())
  offset_slice = 0

  for batch in batches:
    if load_seqs:
      dataset.load_seqs(batch.start_seq, batch.end_seq)
    device.num_frames += batch.get_total_num_frames()
    with dataset.lock:
      for seq in batch.seqs:
        o = seq.batch_frame_offset
        q = seq.batch_slice + offset_slice
        length = seq.frame_length
        # input-data, input-index will also be set in this loop. That is data-key "data".
        # targets are usually data-key "classes".
        for k in device.used_data_keys:
          # device.used_data_keys are set by the train-net, but we will also get here during forward-only,
          # e.g. via SprintInterface, where we don't have e.g. the "classes" data.
          # In that case, l.get(k) should be None. In some earlier code, l.get(k) could also be 0 in that case.
          if length.get(k) in [0, None]:
            continue
          data = dataset.get_data_slice(seq.seq_idx, k, seq.seq_start_frame[k], seq.seq_end_frame[k])
          ls = data.shape[0]
          if "[sparse:" in k:
            assert o[k] == 0, "sparse non-recurrent batching + chunking not implemented"
            _device_maybe_enlarge_data(device, k, ls)
          else:
            if ls != length[k]:
              raise Exception("got shape[0]: %i, expected: %i, start/end: %r/%r, seq_idx: %i, seq len: %r" % (
                ls, length[k], seq.seq_start_frame, seq.seq_end_frame, seq.seq_idx,
                dataset.get_seq_length(seq.seq_idx)))
          device.output_index[k][o[k]:o[k] + ls, q] = numpy.ones((ls,), dtype='int8')
          device.targets[k][o[k]:o[k] + ls, q] = data
        # Only copy ctc targets if chunking is inactive to avoid out of range access.
        # CTC is not compatible with chunking anyway.
        chunking_active = dataset.chunk_size != 0
        if dataset.has_ctc_targets() and not chunking_active:
          device.ctc_targets[q] = dataset.get_ctc_targets(seq.seq_idx)

        device.tags[q] = dataset.get_tag(seq.seq_idx)
    # Note on multiple batches for the non-recurrent case:
    # We could either concatenate all into a single slice, or do multiple slices.
    # We do multiple slices here.
    # See also the `shape` calculation above.
    offset_slice += batch.num_slices

  return True, len(batches)


def _device_maybe_enlarge_data(device, key, needed_len):
  cur_len = device.output_index[key].shape[0]
  if cur_len >= needed_len:
    return
  diff_len = needed_len - cur_len
  new_len = cur_len + int(diff_len * 1.5)  # a bit more than needed
  assert new_len >= needed_len
  # Also see Device.alloc_data() for reference.
  # First, new output_index.
  old_index = device.output_index[key]
  index_shape = list(old_index.shape)
  index_shape[0] = new_len
  device.output_index[key] = numpy.zeros(index_shape, dtype='int8')
  device.output_index[key][0:cur_len] = old_index
  # Now, new targets.
  old_targets = device.targets[key]
  targets_shape = list(old_targets.shape)
  targets_shape[0] = new_len
  device.targets[key] = numpy.full(targets_shape, -1, dtype=device.targets[key].dtype)
  device.targets[key][0:cur_len] = old_targets


def assign_dev_data_single_seq(device, dataset, seq, load_seqs=True):
  """
  :type device: Device.Device
  :type dataset: Dataset.Dataset
  :param int seq: sorted seq idx
  :param bool load_seqs:
  :return: whether we succeeded
  :rtype: bool
  """
  batch = Batch()
  batch.init_with_one_full_sequence(seq_idx=seq, dataset=dataset)
  success, _ = assign_dev_data(device, dataset, [batch], load_seqs=load_seqs)
  return success


def maybe_subtract_priors(network, train, config):
  """
  :type network: Network.LayerNetwork
  :type train: Dataset.Dataset
  :type config: Config.Config
  """
  if config.bool('subtract_priors', False):
    prior_scale = config.float('prior_scale', 0.0)
    priors = train.calculate_priori()
    priors[priors == 0] = 1e-10  # avoid priors of zero which would yield a bias of inf
    ls = [p for p in network.train_params_vars if p.name == 'b_output']
    assert len(ls) == 1, len(ls)
    b_softmax = ls[0]
    b_softmax.set_value(b_softmax.get_value() - prior_scale * numpy.log(priors))
    print("subtracting priors with prior_scale", prior_scale, file=log.v3)




"""
Like SprintErrorSignals.py but for TensorFlow.
"""

from SprintErrorSignals import SprintInstancePool
import tensorflow as tf


def py_get_sprint_automata_for_batch(sprint_opts, tags):
  """
  :param dict[str] sprint_opts:
  :param list[str] tags:
  :return: (edges, weights, start_end_states)
  :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
  """
  # Also see :class:`SprintAlignmentAutomataOp`.
  sprint_instance_pool = SprintInstancePool.get_global_instance(sprint_opts=sprint_opts)
  with sprint_instance_pool.lock:  # We need multi-threading safety.
    edges, weights, start_end_states = sprint_instance_pool.get_automata_for_batch(tags)
  # Note: UnimplementedError: Unsupported numpy type 6 (uint32) -> cast to int32.
  edges = edges.astype("int32")
  start_end_states = start_end_states.astype("int32")
  return edges, weights, start_end_states


def get_sprint_automata_for_batch_op(sprint_opts, tags):
  """
  :param dict[str] sprint_opts:
  :param tf.Tensor tags: shape (batch,), of dtype string
  :return: (edges, weights, start_end_states). all together in one automaton.
    edges are of shape (4, num_edges), each (from, to, emission-idx, seq-idx), of dtype int32.
    weights are of shape (num_edges,), of dtype float32.
    start_end_states are of shape (2, batch), each (start,stop) state idx, batch = len(tags), of dtype int32.
  :rtype: (tf.Tensor, tf.Tensor, tf.Tensor)
  """

  def py_wrap_get_sprint_automata_for_batch(py_tags):
    """
    :param list[str] py_tags: len batch
    :return: (edges, weights, start_end_states)
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    """
    try:
      return py_get_sprint_automata_for_batch(sprint_opts=sprint_opts, tags=py_tags)
    except Exception:
      print("Exception in py_wrap_get_sprint_automata_for_batch:")
      import sys
      sys.excepthook(*sys.exc_info())
      raise

  tags.set_shape((None,))  # (batch,)
  edges, weights, start_end_states = tf.py_func(
    py_wrap_get_sprint_automata_for_batch,
    [tags], [tf.int32, tf.float32, tf.int32],
    name="get_sprint_automata_for_batch")
  assert isinstance(edges, tf.Tensor)
  assert isinstance(weights, tf.Tensor)
  assert isinstance(start_end_states, tf.Tensor)
  edges.set_shape((4, None))  # (4, num_edges)
  weights.set_shape((None,))  # (num_edges,)
  start_end_states.set_shape((2, tags.get_shape().dims[0]))  # (2, batch)
  return edges, weights, start_end_states


def py_get_sprint_loss_and_error_signal(sprint_opts, log_posteriors, seq_lengths, seq_tags):
  """
  :param dict[str] sprint_opts:
  :param numpy.ndarray log_posteriors: 3d (time,batch,label)
  :param numpy.ndarray seq_lengths: 1d (batch)
  :param list[str] seq_tags: seq names
  :return: (loss, error_signal), error_signal has the same shape as posteriors. loss is a 1d-array (batch).
  :rtype: (numpy.ndarray, numpy.ndarray)
  """
  # Also see :class:`SprintErrorSigOp`.
  sprint_instance_pool = SprintInstancePool.get_global_instance(sprint_opts=sprint_opts)
  with sprint_instance_pool.lock:  # We need multi-threading safety.
    loss, error_signal = sprint_instance_pool.get_batch_loss_and_error_signal(
      log_posteriors=log_posteriors, seq_lengths=seq_lengths, tags=seq_tags)
  return loss, error_signal


def get_sprint_loss_and_error_signal(sprint_opts, log_posteriors, seq_lengths, seq_tags):
  """
  :param dict[str] sprint_opts:
  :param tf.Tensor log_posteriors: 3d (time,batch,label)
  :param tf.Tensor seq_lengths: 1d (batch,)
  :param tf.Tensor seq_tags: 1d (batch,), seq names
  :return: (loss, error_signal), error_signal has the same shape as posteriors. loss is a 1d-array (batch).
  :rtype: (tf.Tensor, tf.Tensor)
  """

  def py_wrap_get_sprint_loss_and_error_signal(py_log_posteriors, py_seq_lengths, py_seq_tags):
    """
    :param numpy.ndarray py_log_posteriors: 3d (time,batch,label)
    :param numpy.ndarray py_seq_lengths: 1d (batch)
    :param list[str] py_seq_tags:
    :return: (loss, error_signal), error_signal has the same shape as posteriors. loss is a 1d-array (batch).
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    try:
      return py_get_sprint_loss_and_error_signal(
        sprint_opts=sprint_opts, log_posteriors=py_log_posteriors, seq_lengths=py_seq_lengths, seq_tags=py_seq_tags)
    except Exception:
      print("Exception in py_wrap_get_sprint_loss_and_error_signal:")
      import sys
      sys.excepthook(*sys.exc_info())
      raise

  log_posteriors.set_shape((None, None, None))  # (time,batch,label)
  seq_lengths.set_shape((None,))  # (batch,)
  seq_tags.set_shape((None,))  # (batch,)
  loss, error_signal = tf.py_func(
    py_wrap_get_sprint_loss_and_error_signal,
    [log_posteriors, seq_lengths, seq_tags], [tf.float32, tf.float32],
    name="get_sprint_loss_and_error_signal")
  assert isinstance(loss, tf.Tensor)
  assert isinstance(error_signal, tf.Tensor)
  loss.set_shape((None,))  # (batch,)
  error_signal.set_shape(log_posteriors.get_shape())  # (time,batch,label)
  return loss, error_signal



import numpy as np

import theano

class InvAlignOp(theano.Op):
  __props__ = ('tdps','nstates')

  # index_in, index_out, scores, transcriptions
  itypes = [theano.tensor.bmatrix,theano.tensor.bmatrix,theano.tensor.ftensor3,theano.tensor.imatrix]
  otypes = [theano.tensor.imatrix,theano.tensor.imatrix,theano.tensor.bmatrix]

  # Python implementation:
  def perform(self, node, inputs_storage, output_storage):
    index_in, index_out, scores, transcriptions = inputs_storage[:4]
    shape = (index_out.shape[0] * self.nstates, index_out.shape[1])
    attention = np.zeros(shape, 'int32')
    index = np.zeros(shape, 'int8')
    labelling = np.zeros(shape, 'int32')
    length_x = index_in.sum(axis=0)
    length_y = index_out.sum(axis=0) * self.nstates
    for b in range(scores.shape[1]):
      index[:length_y[b], b] = np.int8(1)
      orth = transcriptions[:length_y[b] / self.nstates, b]
      attention[:length_y[b], b], labelling[:length_y[b], b] = self._viterbi(0, length_x[b], scores[:length_x[b], b], orth)
      attention[:length_y[b], b] += b * index_in.shape[0]
    output_storage[0][0] = labelling
    output_storage[1][0] = attention
    output_storage[2][0] = index
    #print attention

  def __init__(self, tdps, nstates):
    self.nstates = nstates
    self.tdps = tuple(tdps)

  def grad(self, inputs, output_grads):
    return [output_grads[0],output_grads[1],output_grads[2],output_grads[1]]

  def infer_shape(self, node, input_shapes):
    shape = (input_shapes[1][0] * self.nstates, input_shapes[1][1])
    return [shape,shape,shape]

  def _buildHmm(self, transcription):
    start = transcription * self.nstates + 1
    hmm = np.expand_dims(start,axis=1).repeat(self.nstates,axis=1) + np.arange(self.nstates) - 1
    return hmm.astype('int32').flatten()

  def _viterbi(self, start, end, scores, transcription):
    """Fully aligns sequence from start to end but in inverse manner"""
    inf = 1e30
    lengthT = end - start
    skip = max(min(len(self.tdps), lengthT - self.nstates), 1)
    tdps = self.tdps[:skip]
    lengthS = transcription.shape[0] * self.nstates

    hmm = self._buildHmm(transcription)
    # with margins of skip at the bottom or top
    fwdScore = np.full((lengthS, lengthT + skip - 1), inf)
    bt = np.full((lengthS, lengthT + skip - 1), -1, dtype=np.int32)

    # precompute all scores and densities
    score = np.full((lengthS, lengthT + skip - 1), inf)
    for t in range(0, lengthT):
      for s in range(0, lengthS):
        score[s][t + skip - 1] = scores[start + t, hmm[s] / self.nstates]

    # forward
    scores = score[0, 0 + skip - 1:skip + skip - 2]
    scores = np.add(scores, tdps[1:])
    fwdScore[0, 0 + skip - 1:skip + skip - 2] = scores
    bt[0, 0 + skip - 1:skip + skip - 2] = range(1, skip)

    # remaining columns
    for s in range(1, lengthS):
      for t in range(max(lengthT - (lengthS - s) * skip,0), lengthT):
        previous = fwdScore[s - 1, t:t + skip]
        scores = score[s, t + skip - 1]
        scores = np.add(scores, np.add(previous, tdps[::-1]))

        best = np.argmin(scores)
        fwdScore[s, t + skip - 1] = scores[best]
        bt[s, t + skip - 1] = skip - 1 - best

    attention = np.full((lengthS), 0, dtype=np.int32)
    labelling = np.full((lengthS), 0, dtype=np.int32)

    # backtrack
    t = lengthT - 1
    attention[lengthS - 1] = lengthT - 1
    labelling[lengthS - 1] = transcription[-1]
    for s in range(lengthS - 2, -1, -1):
      tnew = t - bt[s + 1][t + skip - 1]
      attention[s] = tnew
      labelling[s] = transcription[s / self.nstates]
      t = tnew
    return attention, labelling

  def _viterbi2(self, start, end, scores, transcription):
    """Fully aligns sequence from start to end but in inverse manner"""
    inf = 1e30
    # max skip transitions derived from tdps

    hmm = self._buildHmm(transcription)
    lengthT = end - start
    skip = max(min(len(self.tdps), lengthT - self.nstates),1)
    tdps = self.tdps[:skip]
    lengthS = transcription.shape[0] * self.nstates

    # with margins of skip at the bottom or top
    fwdScore = np.full((lengthS, lengthT + skip - 1), inf)
    bt = np.full((lengthS, lengthT + skip - 1), -1, dtype=np.int32)

    # precompute all scores and densities
    score = np.full((lengthS, lengthT + skip - 1), inf)
    for t in range(0, lengthT):
      for s in range(0, lengthS):
        score[s][t + skip - 1] = scores[start + t, hmm[s] / self.nstates]

    # forward
    scores = score[0, 0 + skip - 1:skip + skip - 2]
    scores = np.add(scores, tdps[1:])
    fwdScore[0, 0 + skip - 1:skip + skip - 2] = scores
    bt[0, 0 + skip - 1:skip + skip - 2] = range(1, skip)

    # remaining columns
    for s in range(1, lengthS):
      for t in range(max(lengthT - (lengthS - s) * skip,0), lengthT):
        previous = fwdScore[s - 1, t:t + skip]
        scores = score[s, t + skip - 1]
        scores = np.add(scores, np.add(previous, tdps[::-1]))

        best = np.argmin(scores)
        fwdScore[s, t + skip - 1] = scores[best]
        bt[s, t + skip - 1] = skip - 1 - best

    attention = np.full((lengthS), 0, dtype=np.int32)
    labelling = np.full((lengthS), 0, dtype=np.int32)

    # backtrack
    t = lengthT - 1
    attention[lengthS - 1] = lengthT - 1
    labelling[lengthS - 1] = transcription[-1]
    for s in range(lengthS - 2, -1, -1):
      tnew = t - bt[s + 1][t + skip - 1]
      attention[s] = tnew
      labelling[s] = transcription[s / self.nstates]
      t = tnew
    return attention, labelling


class InvFullAlignOp(theano.Op):
  __props__ = ('tdps','nstates')

  # index_in, index_out, scores, transcriptions
  itypes = [theano.tensor.bmatrix,theano.tensor.bmatrix,theano.tensor.ftensor3,theano.tensor.imatrix]
  otypes = [theano.tensor.ftensor3,theano.tensor.bmatrix]

  # Python implementation:
  def perform(self, node, inputs_storage, output_storage):
    index_in, index_out, scores, transcriptions = inputs_storage[:4]
    index = np.zeros((index_out.shape[0] * self.nstates, index_out.shape[1]), 'int8')
    err = np.zeros(scores.shape, 'float32')

    labelling = np.zeros(shape, 'int32')
    length_x = index_in.sum(axis=0)
    length_y = index_out.sum(axis=0) * self.nstates
    for b in range(scores.shape[1]):
      orth = transcriptions[:length_y[b] / self.nstates, b]
      index[:length_y[b], b] = np.int8(1)
      err[:length_x[b],b] = self._baumwelch(0, length_x[b], scores[:length_x[b], b], orth)
    output_storage[0][0] = err
    output_storage[1][0] = index

  def __init__(self, tdps, nstates):
    self.nstates = nstates
    self.tdps = tuple(tdps)

  def grad(self, inputs, output_grads):
    return [output_grads[0],output_grads[1],output_grads[2],output_grads[1]]

  def infer_shape(self, node, input_shapes):
    shape = (input_shapes[1][0] * self.nstates, input_shapes[1][1])
    return [input_shapes[2],(input_shapes[1][0] * self.nstates, input_shapes[1][1])]

  def _buildHmm(self, transcription):
    start = transcription * self.nstates + 1
    hmm = np.expand_dims(start,axis=1).repeat(self.nstates,axis=1) + np.arange(self.nstates) - 1
    return hmm.astype('int32').flatten()

  def _baumwelch(self, start, end, scores, transcription):
    from scipy.misc import logsumexp

    """Fully aligns sequence from start to end but in inverse manner"""
    inf = 1e30
    # max skip transitions derived from tdps

    hmm = self._buildHmm(transcription)
    lengthT = end - start
    skip = min(len(self.tdps), lengthT - self.nstates)
    tdps = self.tdps[:skip]
    lengthS = transcription.shape[0] * self.nstates

    # with margins of skip at the bottom or top
    fwdScore = np.full((lengthS, lengthT + skip - 1), inf)
    bwdScore = np.full((lengthS, lengthT + skip - 1), inf)
    bt = np.full((lengthS, lengthT + skip - 1), -1, dtype=np.int32)

    # precompute all scores and densities
    score = np.full((lengthS, lengthT + skip - 1), inf)
    for t in range(0, lengthT):
      for s in range(0, lengthS):
        score[s][t + skip - 1] = scores[start + t, hmm[s] / self.nstates]

    # forward
    scores = score[0, 0 + skip - 1:skip + skip - 2]
    scores = np.add(scores, tdps[1:])
    fwdScore[0, 0 + skip - 1:skip + skip - 2] = scores
    bt[0, 0 + skip - 1:skip + skip - 2] = range(1, skip)

    # remaining columns
    for s in range(1, lengthS):
      for t in range(max(lengthT - (lengthS - s) * skip,0), lengthT - (lengthS - s - 1)):
        previous = fwdScore[s - 1, t:t + skip]
        scores = score[s, t + skip - 1]
        scores = np.add(scores, np.add(previous, tdps[::-1]))
        fwdScore[s, t + skip - 1] = -logsumexp(-scores)

    score = score[:, skip - 1:lengthT + skip - 1]
    score = np.pad(score, pad_width=((0, 0), (0, skip - 1)), mode='constant',
                   constant_values=inf)

    # backward
    # initalize last columns, use lm of last character
    bwdScore[lengthS - 1][lengthT - 1] = score[lengthS - 1][lengthT - 1]
    # remaining columns
    for s in range(lengthS - 2, -1, -1):
      for t in range(0, lengthT):
        previous = bwdScore[s + 1, t:t + skip]
        scores = np.add(score[s, t], previous)
        bwdScore[s, t] = -logsumexp(-np.add(scores, np.asarray(self.tdps)))

    errScore = np.add(fwdScore, bwdScore)
    errScore = -np.logaddexp(-errScore, logsumexp(-errScore,axis=0))

    return errScore


class InvBacktrackOp(theano.Op):
  # Properties attribute
  __props__ = ('tdps', 'nstates', "penalty")

  # index_in, scores
  itypes = [theano.tensor.bmatrix, theano.tensor.ftensor3, theano.tensor.ftensor3]
  otypes = [theano.tensor.imatrix, theano.tensor.imatrix, theano.tensor.bmatrix]

  # Python implementation:
  def perform(self, node, inputs_storage, output_storage):
    index_in, scores, skips = inputs_storage[:3]
    transcript = np.zeros(index_in.shape,'int32')
    attention = np.zeros(index_in.shape, 'int32')
    index = np.zeros(index_in.shape, 'int8')
    for b in range(scores.shape[1]):
      length_x = index_in[:,b].sum()
      t, a = self._recognize(scores[:length_x, b], skips[:length_x, b])
      #if b == 0:
      #  print np.exp(-transitions[:length_x, b])
      length_y = len(a)
      transcript[:length_y, b] = t
      attention[:length_y, b] = np.asarray(a) + b * index_in.shape[0]
      index[:length_y, b] = 1

    output_storage[0][0] = transcript
    output_storage[1][0] = attention
    output_storage[2][0] = index

  def __init__(self, tdps, nstates, penalty):
    self.nstates = nstates
    self.penalty = penalty
    self.tdps = tuple(tdps)

  def grad(self, inputs, output_grads):
    return [output_grads[0], output_grads[1], output_grads[2]]

  def infer_shape(self, node, input_shapes):
    return [input_shapes[0], input_shapes[0], input_shapes[0]]

  def _buildHmm(self, transcription):
    transcriptionLength = transcription.shape[0]
    hmm = np.zeros(transcriptionLength * self.nstates, dtype=np.int32)

    for i in range(0, transcriptionLength):
      startState = transcription[i] * self.nstates + 1
      for s in range(0, self.nstates):
        hmm[i * self.nstates + s] = startState + s - 1

    return hmm

  def _recognize2(self, scores, transitions):
    lengthT = scores.shape[0]
    transcript = []
    attention = []
    t = lengthT - 1
    while t >= 0:
      label = scores[t].argmin()
      if label % self.nstates == self.nstates - 1:
        if True or len(transcript) == 0 or label / self.nstates != transcript[-1]:
          attention.append(t)
          transcript.append(label / self.nstates)
      t -= transitions[t].argmin() + 1
    return transcript[::-1], attention[::-1]

  def _recognize(self, scores, skips):
    lengthT = scores.shape[0]
    transcript = []
    attention = []
    t = lengthT - 1
    n = 0
    m = lengthT
    lsc = 0
    #print skips.argmin(axis=1)
    while t >= 0:
      label = scores[t].argmin()
      lsc += scores[t]
      if n % self.nstates == 0:
        attention.append(t)
        #transcript.append(np.sum(scores[t:m]).argmin())
        #m = t
        transcript.append(lsc.argmin())
        lsc = 0
      t -= (skips[t,1:].argmin() + 1) #* 3 + 1
      n += 1
    return transcript[::-1], attention[::-1]

  def _recognize3(self, scores, transitions):
    lengthT = scores.shape[0]
    lengthS = transitions.shape[1]

    cost = np.full((lengthT, lengthT), np.inf, 'float32')
    back = np.full((lengthT, lengthT), np.inf, 'int32')

    cost[0] = np.min(scores[0])
    back[0] = -1

    transcript = []
    attention = []

    for s in xrange(1, lengthT):
      for t in xrange(min(s * lengthS, lengthT)):
        #if s % self.nstates == 0: # end state

        cost[s, t] = np.min(scores[s])
        q = transitions[t].copy()
        q[:min(t,lengthS)] += cost[s - 1, t - min(t,lengthS) : t]
        back[s, t] = q.argmin() + 1
        cost[s, t] += q.min()

    t = lengthT - 1
    s = 1
    while t >= 0 and s < lengthT:
      if s % self.nstates == 0:
        attention.append(t)
        transcript.append(scores[t].argmin()  / self.nstates)
      t -= back[-s, t]
      s += 1
    return transcript[::-1], attention[::-1]


class InvDecodeOp(theano.Op):
  # Properties attribute
  __props__ = ('tdps', 'nstates', "penalty")

  # index_in, scores
  itypes = [theano.tensor.bmatrix, theano.tensor.ftensor3]
  otypes = [theano.tensor.imatrix, theano.tensor.imatrix, theano.tensor.bmatrix]

  # Python implementation:
  def perform(self, node, inputs_storage, output_storage):
    index_in, scores = inputs_storage[:2]
    transcript = np.zeros(index_in.shape, 'int32')
    attention = np.zeros(index_in.shape, 'int32')
    index = np.zeros(index_in.shape, 'int8')
    max_length_y = 0
    for b in range(scores.shape[1]):
      length_x = index_in[:, b].sum()
      t, a = self._recognize(0, length_x, scores[:length_x, b])
      length_y = len(a)
      transcript[:length_y, b] = t
      attention[:length_y, b] = np.asarray(a) + b * index_in.shape[0]
      index[:length_y, b] = 1
      max_length_y = max(length_y, max_length_y)

    output_storage[0][0] = transcript
    output_storage[1][0] = attention
    output_storage[2][0] = index

  def __init__(self, tdps, nstates, penalty):
    self.nstates = nstates
    self.penalty = penalty
    self.tdps = tuple(tdps)

  def grad(self, inputs, output_grads):
    return [output_grads[0] * 0, output_grads[1] * 0, output_grads[2] * 0]

  def infer_shape(self, node, input_shapes):
    return [input_shapes[0], input_shapes[0], input_shapes[0]]

  def _buildHmm(self, transcription):
    transcriptionLength = transcription.shape[0]
    hmm = np.zeros(transcriptionLength * self.nstates, dtype=np.int32)

    for i in range(0, transcriptionLength):
      startState = transcription[i] * self.nstates + 1
      for s in range(0, self.nstates):
        hmm[i * self.nstates + s] = startState + s - 1

    return hmm

  def _recognize(self, start, end, scores):
    """recognizes a sequence from start until (excluded) end
    with inverse search s -> t_s"""
    inf = 1e30
    bestAttend = []
    bestResult = []
    bestScore = inf

    skip = len(self.tdps)
    maxDuration = 0.8
    mod_factor = 1.0

    hmmLength = scores.shape[1] * self.nstates
    seqLength = end - start
    leftScore = np.full((hmmLength, seqLength + skip - 1), inf)
    rightScore = np.full((hmmLength, seqLength + skip - 1), inf)

    leftStart = np.full((hmmLength, seqLength + skip - 1), 0, dtype=np.uint64)
    rightStart = np.full((hmmLength, seqLength + skip - 1), 0, dtype=np.uint64)

    leftEpoch = np.full((hmmLength, seqLength + skip - 1), - 1, dtype=np.int64)
    rightEpoch = np.full((hmmLength, seqLength + skip - 1), - 1, dtype=np.int64)

    bestWordEnds = np.full((int(maxDuration * seqLength), seqLength), - 1, dtype=np.int64)
    bestWordEndsStart = np.full((int(maxDuration * seqLength), seqLength), - 1, dtype=np.int64)
    bestWordEndsEpoch = np.full((int(maxDuration * seqLength), seqLength), - 1, dtype=np.int64)

    # precompute all scores and densities
    score = np.full((hmmLength, seqLength + skip), inf)
    for t in range(0, seqLength):
      for s in range(0, hmmLength):
        score[s][t + skip - 1] = scores[start + t, s / self.nstates]

    # apply model scale
    score = np.multiply(score, mod_factor)

    # initalize for first timeframe
    leftScore[0][skip - 1] = score[0][skip - 1]
    leftEpoch[0][skip - 1] = 0
    bestWordEnds[0][0] = 0
    bestWordEndsStart[0][0] = 0
    bestWordEndsEpoch[0][0] = 0

    for s in range(1, hmmLength, self.nstates):
      leftScore[s][skip - 1] = score[s][skip - 1]
      leftEpoch[s][skip - 1] = 0

    for a in range(1, int(maxDuration * seqLength)):
      # determine best word ends and score
      bestWordEnd = np.argmin(np.concatenate(
        ([leftScore[0]],
         leftScore[self.nstates:hmmLength:
         self.nstates] + self.penalty)), axis=0)

      bestWordEnd = bestWordEnd * self.nstates

      # two times min might be improved
      bestWordEndScore = np.min(np.concatenate(
        ([leftScore[0]],
         leftScore[self.nstates:hmmLength:
         self.nstates] + self.penalty)), axis=0)

      bestWordEnds[a] = bestWordEnd[skip - 1::]
      for t in range(0, seqLength):
        bestWordEndsStart[a][t] = leftStart[bestWordEnd[t + skip - 1]][t + skip - 1]
        bestWordEndsEpoch[a][t] = leftEpoch[bestWordEnd[t + skip - 1]][t + skip - 1]

      for s in range(0, hmmLength):
        # for silence allow special transitions (only 1)
        if s == 0:
          for t in range(0, seqLength - 1):
            if leftScore[0][t + skip - 1] < bestWordEndScore[t + skip - 1] or bestWordEnd[t + skip - 1] == 0:
              # inner silence
              rightScore[0][t + skip] = leftScore[0][t + skip - 1] + score[0][t + skip]
              rightStart[0][t + skip] = leftStart[0][t + skip - 1]
              rightEpoch[0][t + skip] = leftEpoch[0][t + skip - 1]
            else:
              # to silence
              # only transition 1 is allowed
              rightScore[0][t + skip] = bestWordEndScore[t + skip - 1] + score[0][t + skip]
              rightStart[0][t + skip] = t + 1
              rightEpoch[0][t + skip] = a

        # between word transitions
        elif s % self.nstates == 1:
          for t in range(0, seqLength):
            # linear interpolate scores
            scores = np.add(np.add(
              np.cumsum(np.append(
                [0.], score[s, t + 1:t + skip][::-1])), self.tdps),
              bestWordEndScore[t:t + skip][::-1])

            # index corresponds to transition
            bestChoice = np.argmin(scores)

            rightScore[s][t + skip - 1] = scores[bestChoice]
            rightStart[s][t + skip - 1] = t + 1 - bestChoice
            rightEpoch[s][t + skip - 1] = a

        # inner word transitions
        else:
          for t in range(0, seqLength):
            # linear interpolate scores
            scores = np.add(np.add(
              np.cumsum(np.append(
                [0.], score[s, t + 1:t + skip][::-1])), self.tdps),
              leftScore[s - 1, t:t + skip][::-1])

            # index corresponds to transition
            bestChoice = np.argmin(scores)

            rightScore[s][t + skip - 1] = scores[bestChoice]
            rightStart[s][t + skip - 1] = leftStart[s - 1][t + skip - 1 - bestChoice]
            rightEpoch[s][t + skip - 1] = leftEpoch[s - 1][t + skip - 1 - bestChoice]

      # finish epoch
      leftScore, rightScore = rightScore, leftScore
      leftStart, rightStart = rightStart, leftStart
      leftEpoch, rightEpoch = rightEpoch, leftEpoch

      rightScore = np.full((hmmLength, seqLength + skip - 1), inf)
      rightStart = np.full((hmmLength, seqLength + skip - 1), start, dtype=np.uint64)
      rightEpoch = np.full((hmmLength, seqLength + skip - 1), - 1, dtype=np.int64)

      # backtrack
      # backtrace = []
      if bestWordEndScore[-1] < inf:
        result = []
        attend = []
        t_idx = seqLength - 1
        a_idx = a
        while t_idx > 0:
          result.append(a_idx / self.nstates)
          attend.append(t_idx)

          # backtrace.append((result[-1],
          #     t_idx, bestWordEndsStart[a_idx][t_idx] - 1,
          #     a_idx, bestWordEndsEpoch[a_idx][t_idx]))
          t_idx_temp = bestWordEndsStart[a_idx][t_idx] - 1
          a_idx = bestWordEndsEpoch[a_idx][t_idx]
          t_idx = t_idx_temp

        if bestWordEndScore[-1] < bestScore:
          bestResult = list(reversed(result))
          bestScore = bestWordEndScore[-1]
          bestAttend = list(reversed(attend))
    return bestResult, bestAttend



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author: Pavel Golik (golik@cs.rwth-aachen.de)

"""
This module is about reading (maybe later also writing) the Sprint archive format.
"""

from __future__ import print_function

import sys
import os
import typing
import array
from struct import pack, unpack
import numpy
import zlib
import mmap


class FileInfo:
  """
  File info.
  """

  def __init__(self, name, pos, size, compressed, index):
    """
    :param str name:
    :param int pos:
    :param int size:
    :param bool|int compressed:
    :param int index:
    """
    self.name = name
    self.pos = pos
    self.size = size
    self.compressed = compressed
    self.index = index

  def __repr__(self):
    return "FileInfo(%s)" % " ".join(str(s) for s in self.__dict__.values())


class FileArchive:
  """
  File archive.
  """

  # read routines
  def read_u32(self):
    """
    :rtype: int
    """
    return int(unpack("i", self.f.read(4))[0])

  # noinspection PyPep8Naming
  def read_U32(self):
    """
    :rtype: int
    """
    return int(unpack("I", self.f.read(4))[0])

  def read_u64(self):
    """
    :rtype: int
    """
    return int(unpack("q", self.f.read(8))[0])

  def read_char(self):
    """
    :rtype: int
    """
    return unpack("b", self.f.read(1))[0]

  def read_bytes(self, l):
    """
    :rtype: bytes
    """
    return unpack('%ds' % l, self.f.read(l))[0]

  def read_str(self, l, enc='ascii'):
    """
    :rtype: str
    """
    return self.read_bytes(l).decode(enc)

  def read_f32(self):
    """
    :rtype: float
    """
    return float(unpack("f", self.f.read(4))[0])

  def read_f64(self):
    """
    :rtype: float
    """
    return float(unpack("d", self.f.read(8))[0])

  def read_v(self, typ, size):
    """
    :param str typ: "f" for float (float32) or "d" for double (float64)
    :param int size: number of elements to return
    :return: numpy array of shape (size,) of dtype depending on typ
    :rtype: numpy.ndarray
    """
    if typ == 'f':
      b = 4
      t = numpy.float32
    elif typ == 'd':
      b = 8
      t = numpy.float64
    else:
      raise NotImplementedError("typ: %r" % typ)
    if isinstance(self.f, mmap.mmap):
      res = numpy.frombuffer(self.f, t, size, self.f.tell())
      self.f.seek(b * size, os.SEEK_CUR)
    else:
      res = numpy.fromfile(self.f, t, size, '')
    return res

  # write routines
  def write_str(self, s):
    """
    :param str s:
    :rtype: int
    """
    return self.f.write(pack("%ds" % len(s), s))

  def write_char(self, i):
    """
    :param int i:
    :rtype: int
    """
    return self.f.write(pack("b", i))

  def write_u32(self, i):
    """
    :param int i:
    :rtype: int
    """
    return self.f.write(pack("i", i))

  # noinspection PyPep8Naming
  def write_U32(self, i):
    """
    :param int i:
    :rtype: int
    """
    return self.f.write(pack("I", i))

  def write_u64(self, i):
    """
    :param int i:
    :rtype: int
    """
    return self.f.write(pack("q", i))

  def write_f32(self, i):
    """
    :param float i:
    :rtype: int
    """
    return self.f.write(pack("f", i))

  def write_f64(self, i):
    """
    :param float i:
    :rtype: int
    """
    return self.f.write(pack("d", i))

  SprintCacheHeader = "SP_ARC1\0"
  start_recovery_tag = 0xaa55aa55
  end_recovery_tag = 0x55aa55aa

  def __init__(self, filename, must_exists=True):

    self.ft = {}  # type: typing.Dict[str,FileInfo]
    if os.path.exists(filename):
      self.allophones = []
      self.f = open(filename, 'rb')
      header = self.read_str(len(self.SprintCacheHeader))
      assert header == self.SprintCacheHeader

      ft = bool(self.read_char())
      if ft:
        self.read_file_info_table()
      else:
        self.scan_archive()

    else:
      assert not must_exists, "File does not exist: %r" % filename
      self.f = open(filename, 'wb')
      self.write_str(self.SprintCacheHeader)
      self.write_char(1)

    self._short_seg_names = {os.path.basename(n): n for n in self.ft.keys()}
    if len(self._short_seg_names) < len(self.ft):
      # We don't have a unique mapping, so we cannot use this.
      self._short_seg_names.clear()

  def __del__(self):
    self.f.close()

  def file_list(self):
    """
    :rtype: list[str]
    """
    return self.ft.keys()

  def finalize(self):
    """
    Finalize.
    """
    self.write_file_info_table()

  def read_file_info_table(self):
    """
    Read file info table.
    """
    self.f.seek(-8, 2)
    pos_count = self.read_u64()
    self.f.seek(pos_count)
    count = self.read_u32()
    if not count > 0:
      return
    for i in range(count):
      str_len = self.read_u32()
      name = self.read_str(str_len)
      pos = self.read_u64()
      size = self.read_u32()
      comp = self.read_u32()
      self.ft[name] = FileInfo(name, pos, size, comp, i)
      # TODO: read empty files

  def write_file_info_table(self):
    """
    Write file info table.
    """
    pos = self.f.tell()
    self.write_u32(len(self.ft))

    for fi in self.ft.values():
      self.write_u32(len(fi.name))
      self.write_str(fi.name)
      self.write_u64(fi.pos)
      self.write_u32(fi.size)
      self.write_u32(fi.compressed)

    self.write_u64(0)
    self.write_u64(pos)

  def scan_archive(self):
    """
    Scan archive.
    """
    i = 0
    self.f.seek(0, 2)
    end = self.f.tell()
    self.f.seek(0)
    while self.f.tell() < end:
      if self.read_U32() != self.start_recovery_tag:
        continue

      fn_len = self.read_u32()
      name = self.read_str(fn_len)
      pos = self.f.tell()
      size = self.read_u32()
      comp = self.read_u32()
      self.read_u32()  # chk

      self.f.seek(size, 1)
      self.ft[name] = FileInfo(name, pos, size, comp, i)
      i += 1

      self.read_U32()  # end tag

      # raise NotImplementedError("Need to scan archive if no "
      #                           "file info table found.")

  def _raw_read(self, size, typ):
    """
    :param int|None size: needed for typ == "str"
    :param str typ: "str", "feat" or "align"
    :return: depending on typ, "str" -> string, "feat" -> (time, data), "align" -> align,
      where string is a str,
      time is list of time-stamp tuples (start-time,end-time) in millisecs,
        data is a list of features, each a numpy vector,
      align is a list of (time, allophone, state), time is an int from 0 to len of align,
        allophone is some int, state is e.g. in [0,1,2].
    :rtype: str|(list[numpy.ndarray],list[numpy.ndarray])|list[(int,int,int)]
    """

    if typ == "str":
      return self.read_str(size)

    elif typ == "feat":
      type_len = self.read_U32()
      typ = self.read_str(type_len)
      # print(typ)
      assert typ == "vector-f32"
      count = self.read_U32()
      data = [None] * count  # type: typing.List[typing.Optional[numpy.ndarray]]
      time_ = [None] * count  # type: typing.List[typing.Optional[numpy.ndarray]]
      for i in range(count):
        size = self.read_U32()
        data[i] = self.read_v("f", size)  # size x f32
        time_[i] = self.read_v("d", 2)  # 2    x f64
      return time_, data

    elif typ in ["align", "align_raw"]:
      type_len = self.read_U32()
      typ = self.read_str(type_len)
      assert typ == "flow-alignment"
      self.read_u32()  # flag ?
      typ = self.read_str(8)
      if typ in ["ALIGNRLE", "AALPHRLE"]:
        # In case of AALPHRLE, after the alignment, we include the alphabet of the used labels.
        # We ignore this at the moment.
        size = self.read_U32()
        if size < (1 << 31):
          # RLE scheme
          time = 0
          alignment = []
          while len(alignment) < size:
            n = self.read_char()
            # print(n)
            if n > 0:
              while n > 0:
                mix, state = self.read_u32(), None
                if typ != "align_raw":
                  mix, state = self.get_state(mix)
                # print(mix, state)
                # print(time, self.allophones[mix])
                alignment.append((time, mix, state))
                time += 1
                n -= 1
            elif n < 0:
              mix, state = self.read_u32(), None
              if typ != "align_raw":
                mix, state = self.get_state(mix)
              while n < 0:
                # print(mix, state)
                # print(time, self.allophones[mix])
                alignment.append((time, mix, state))
                time += 1
                n += 1
            else:
              time = self.read_u32()
              # print("time", time)
          return alignment
        else:
          raise NotImplementedError("No support for weighted "
                                    "alignments yet.")
      else:
        raise Exception("No valid alignment header found (found: %r). Wrong cache?" % typ)

  def has_entry(self, filename):
    """
    :param str filename: argument for self.read()
    :return: True if we have this entry
    """
    return filename in self.ft

  def read(self, filename, typ):
    """
    :param str filename: the entry-name in the archive
    :param str typ: "str", "feat" or "align"
    :return: depending on typ, "str" -> string, "feat" -> (time, data), "align" -> align,
      where string is a str,
      time is list of time-stamp tuples (start-time,end-time) in millisecs,
        data is a list of features, each a numpy vector,
      align is a list of (time, allophone, state), time is an int from 0 to len of align,
        allophone is some int, state is e.g. in [0,1,2].
    :rtype: str|(list[numpy.ndarray],list[numpy.ndarray])|list[(int,int,int)]
    """

    if filename not in self.ft:
      if filename in self._short_seg_names:
        filename = self._short_seg_names[filename]

    fi = self.ft[filename]
    self.f.seek(fi.pos)
    size = self.read_U32()
    comp = self.read_U32()
    self.read_U32()  # chk
    if size == 0:
      return None

    if comp > 0:
      # read compressed bytes into memory as 'bytearray'
      a = array.array('b')
      a.fromfile(self.f, comp)
      # unpack
      b = zlib.decompress(a.tostring(), 15+32)
      # substitute self.f by an anonymous memmap file object
      # restore original file handle after we're done
      backup_f = self.f
      self.f = mmap.mmap(-1, len(b))
      self.f.write(b)
      self.f.seek(0)
      try:
        return self._raw_read(size=fi.size, typ=typ)
      finally:
        self.f = backup_f

    return self._raw_read(size=fi.size, typ=typ)

  def get_state(self, mix):
    """
    :param int mix:
    :return: (mix, state)
    :rtype: (int,int)
    """
    # See src/Tools/Archiver/Archiver.cc:getStateInfo() from Sprint source code.
    assert self.allophones
    max_states = 6
    state = None
    for state in range(max_states):
      if mix >= len(self.allophones):
        mix -= (1 << 26)
      else:
        break
    assert mix >= 0
    return mix, state

  def set_allophones(self, f):
    """
    :param str f: allophone filename. line-separated. will ignore lines starting with "#"
    """
    del self.allophones[:]
    for line in open(f):
      line = line.strip()
      if line.startswith("#"):
        continue
      self.allophones.append(line)

  def add_feature_cache(self, filename, features, times):
    """
    :param str filename:
    :param features:
    :param times:
    """
    self.write_U32(self.start_recovery_tag)
    self.write_u32(len(filename))
    self.write_str(filename)
    pos = self.f.tell()
    size = 4 + 10 + 4 + len(features) * (4 + len(features[0]) * 4 + 2 * 8)
    self.write_u32(size)
    self.write_u32(0)
    self.write_u32(0)

    self.write_u32(10)
    self.write_str("vector-f32")
    assert len(features) == len(times)
    self.write_u32(len(features))
    for f, t in zip(features, times):
      self.write_u32(len(f))
      for x in f.flat:
        self.write_f32(x)
      self.write_f64(t[0])
      self.write_f64(t[1])

    self.ft[filename] = FileInfo(filename, pos, size, 0, len(self.ft))
    self.write_U32(self.end_recovery_tag)

    self.add_attributes(filename, len(features[0]), times[-1][1])

  def add_attributes(self, filename, dim, duration):
    """
    :param str filename:
    :param int dim:
    :param float duration:
    """
    data = ('<flow-attributes>'
            '<flow-attribute name="datatype" value="vector-f32"/>'
            '<flow-attribute name="sample-size" value="%d"/>'
            '<flow-attribute name="total-duration" value="%.5f"/>'
            '</flow-attributes>') % (dim, duration)
    self.write_U32(self.start_recovery_tag)
    filename = "%s.attribs" % filename
    self.write_u32(len(filename))
    self.write_str(filename)
    pos = self.f.tell()
    size = len(data)
    self.write_u32(size)
    self.write_u32(0)
    self.write_u32(0)
    self.write_str(data)
    self.write_U32(self.end_recovery_tag)
    self.ft[filename] = FileInfo(filename, pos, size, 0, len(self.ft))


class FileArchiveBundle:
  """
  File archive bundle.
  """

  def __init__(self, filename):
    """
    :param str filename: .bundle file
    """
    # filename -> FileArchive
    self.archives = {}  # type: typing.Dict[str,FileArchive]
    # archive content file -> FileArchive
    self.files = {}  # type: typing.Dict[str,FileArchive]
    self._short_seg_names = {}
    for line in open(filename).read().splitlines():
      self.archives[line] = a = FileArchive(line, must_exists=True)
      for f in a.ft.keys():
        self.files[f] = a
      # noinspection PyProtectedMember
      self._short_seg_names.update(a._short_seg_names)

  def file_list(self):
    """
    :rtype: list[str]
    :returns: list of content-filenames (which can be used for self.read())
    """
    return self.files.keys()

  def has_entry(self, filename):
    """
    :param str filename: argument for self.read()
    :return: True if we have this entry
    """
    return filename in self.files

  def read(self, filename, typ):
    """
    :param str filename: the entry-name in the archive
    :param str typ: "str", "feat" or "align"
    :return: depending on typ, "str" -> string, "feat" -> (time, data), "align" -> align,
      where string is a str,
      time is list of time-stamp tuples (start-time,end-time) in millisecs,
        data is a list of features, each a numpy vector,
      align is a list of (time, allophone, state), time is an int from 0 to len of align,
        allophone is some int, state is e.g. in [0,1,2].
    :rtype: str|(list[numpy.ndarray],list[numpy.ndarray])|list[(int,int,int)]

    Uses FileArchive.read().
    """
    if filename not in self.files:
      if filename in self._short_seg_names:
        filename = self._short_seg_names[filename]
    return self.files[filename].read(filename, typ)

  def set_allophones(self, filename):
    """
    :param str filename: allophone filename
    """
    for a in self.archives.values():
      a.set_allophones(filename)


def open_file_archive(archive_filename, must_exists=True):
  """
  :param str archive_filename:
  :param bool must_exists:
  :rtype: FileArchiveBundle|FileArchive
  """
  if archive_filename.endswith(".bundle"):
    assert must_exists
    return FileArchiveBundle(archive_filename)
  else:
    return FileArchive(archive_filename, must_exists=must_exists)


def is_sprint_cache_file(filename):
  """
  :param str filename: file to check. must exist
  :return: True iff this is a sprint cache (which can be loaded with `open_file_archive()`)
  :rtype: bool
  """
  assert os.path.exists(filename)
  if not os.path.isfile(filename):
    return False
  if filename.endswith(".bundle"):
    return True
  with open(filename, 'rb') as f:
    header_len = len(FileArchive.SprintCacheHeader)
    bs = unpack('%ds' % header_len, f.read(header_len))[0]
    header = bs.decode("ascii")
    return header == FileArchive.SprintCacheHeader


class AllophoneLabeling(object):
  """
  Allophone labeling.
  """

  def __init__(self, silence_phone, allophone_file, phoneme_file=None, state_tying_file=None, verbose_out=None):
    """
    :param str silence_phone: e.g. "si"
    :param str allophone_file: list of allophones
    :param str|None phoneme_file: list of phonemes
    :param str|None state_tying_file: allophone state tying (e.g. via CART). maps each allophone state to a class label
    :param file verbose_out: stream to dump log messages
    """
    assert phoneme_file or state_tying_file
    self.allophone_file = allophone_file
    self.allophones = [l for l in open(allophone_file).read().splitlines() if l and l[0] != "#"]
    self.allophones_idx = {p: i for i, p in enumerate(self.allophones)}
    self.sil_allo_state_id = self.allophones_idx[silence_phone + "{#+#}@i@f"]
    if verbose_out:
      print("AllophoneLabeling: Num allophones: %i" % len(self.allophones), file=verbose_out)
    self.sil_label_idx = None
    self.num_labels = None
    self.phonemes = None
    self.phoneme_idxs = None
    self.state_tying = None
    self.state_tying_by_allo_state_idx = None
    self.num_allo_states = None
    if phoneme_file:
      self.phonemes = open(phoneme_file).read().splitlines()
      self.phoneme_idxs = {p: i for i, p in enumerate(self.phonemes)}
      if not state_tying_file:
        self.sil_label_idx = self.phoneme_idxs[silence_phone]
        self.num_labels = len(self.phonemes)
        if verbose_out:
          print("AllophoneLabeling: %i phones = labels." % self.num_labels, file=verbose_out)
    if state_tying_file:
      self.state_tying = {k: int(v)
                          for l in open(state_tying_file).read().splitlines()
                          for (k, v) in [l.split()]}
      self.sil_label_idx = self.state_tying[silence_phone + "{#+#}@i@f.0"]
      self.num_allo_states = self._get_num_allo_states()
      self.state_tying_by_allo_state_idx = {
        a + s * (1 << 26): self.state_tying["%s.%i" % (a_s, s)]
        for (a, a_s) in enumerate(self.allophones)
        for s in range(self.num_allo_states)
        if ("%s.%i" % (a_s, s)) in self.state_tying}
      self.num_labels = max(self.state_tying.values()) + 1
      if verbose_out:
        print("AllophoneLabeling: State tying with %i labels." % self.num_labels, file=verbose_out)
    assert self.num_labels is not None
    assert self.state_tying or self.phoneme_idxs

  def _get_num_allo_states(self):
    assert self.state_tying
    return max([int(s.split(".")[-1]) for s in self.state_tying.keys()]) + 1

  def get_label_idx_by_allo_state_idx(self, allo_state_idx):
    """
    :param int allo_state_idx:
    :rtype: int
    """
    if self.state_tying_by_allo_state_idx:
      return self.state_tying_by_allo_state_idx[allo_state_idx]
    # See getState above().
    max_states = 6
    allo_idx = allo_state_idx
    state_idx = 0
    for state_idx in range(max_states):
      if allo_idx >= len(self.allophones):
        allo_idx -= (1 << 26)
      else:
        break
    assert allo_idx >= 0
    return self.get_label_idx(allo_idx, state_idx)

  def get_label_idx(self, allo_idx, state_idx):
    """
    :param int allo_idx:
    :param int state_idx:
    :rtype: int
    """
    if self.state_tying_by_allo_state_idx:
      try:
        return self.state_tying_by_allo_state_idx[allo_idx + state_idx * (1 << 26)]
      except KeyError:
        allo_str = self.allophones[allo_idx]
        r = self.state_tying.get("%s.%i" % (allo_str, state_idx))
        raise KeyError("allo idx %i (%r), state idx %i not found; entry: %r" % (allo_idx, allo_str, state_idx, r))
    allo_str = self.allophones[allo_idx]
    phone = allo_str[:allo_str.index("{")]
    return self.phoneme_idxs[phone]


###############################################################################

class MixtureSet:
  """
  Mixture set.
  """

  # read routines
  def read_u32(self):
    """
    :rtype: int
    """
    return int(unpack('i', self.f.read(4))[0])

  # noinspection PyPep8Naming
  def read_U32(self):
    """
    :rtype: int
    """
    return int(unpack('I', self.f.read(4))[0])

  def read_u64(self):
    """
    :rtype: int
    """
    return int(unpack('q', self.f.read(8))[0])

  def read_char(self):
    """
    :rtype: int
    """
    return unpack('b', self.f.read(1))[0]

  def read_str(self, l, enc='ascii'):
    """
    :param int l:
    :param str enc:
    :rtype: str
    """
    a = array.array('b')
    a.fromfile(self.f, l)
    return a.tostring().decode(enc)

  def read_f32(self):
    """
    :rtype: float
    """
    return float(unpack('f', self.f.read(4))[0])

  def read_f64(self):
    """
    :rtype: float
    """
    return float(unpack('d', self.f.read(8))[0])

  def read_v(self, size, a):
    """
    :param int size:
    :param array.array a:
    :rtype: array.array
    """
    a.fromfile(self.f, size)
    return a

  def __init__(self, filename):
    """
    :param str filename:
    """
    self.header = 'MIXSET\0'
    self.f = open(filename, 'rb')
    header = self.read_str(8)
    assert header[:7] == self.header
    self.version = self.read_u32()
    self.dim = self.read_u32()

    self.num_means = self.read_u32()
    self.means = numpy.zeros([self.num_means, self.dim], dtype=numpy.float64)
    self.mean_weights = numpy.zeros(self.num_means, dtype=numpy.float64)

    for n in range(self.num_means):
      size = self.read_u32()
      assert size == self.dim
      arr_f = array.array('d')
      arr_f.fromfile(self.f, self.dim)
      self.means[n, :] = numpy.array(arr_f)
      self.mean_weights[n] = self.read_f64()

    self.num_covs = self.read_u32()
    self.covs = numpy.zeros([self.num_covs, self.dim], dtype=numpy.float64)
    self.cov_weights = numpy.zeros(self.num_covs, dtype=numpy.float64)

    for n in range(self.num_covs):
      size = self.read_u32()
      assert size == self.dim
      arr_f = array.array('d')
      arr_f.fromfile(self.f, self.dim)
      self.covs[n, :] = numpy.array(arr_f)
      self.cov_weights[n] = self.read_f64()

    self.num_densities = self.read_u32()
    self.densities = numpy.zeros((self.num_densities, 2), dtype=numpy.int32)
    for n in range(self.num_densities):
      mean_idx = self.read_u32()
      cov_idx = self.read_u32()
      self.densities[n, 0] = mean_idx
      self.densities[n, 1] = cov_idx

    self.num_mixtures = self.read_u32()
    self.mixtures = [None] * self.num_mixtures  # type: typing.List[typing.Optional[typing.Tuple[typing.List[int],typing.List[float]]]]  # nopep8
    for n in range(self.num_mixtures):
      num_densities = self.read_u32()
      dns_idx = []
      dns_weight = []
      for i in range(num_densities):
        dns_idx.append(self.read_u32())
        dns_weight.append(self.read_f64())
      self.mixtures[n] = (dns_idx, dns_weight)

  # TODO?
  # noinspection PyUnresolvedReferences
  def write(self, filename):
    """
    :param str filename:
    """
    self.f = open(filename, 'wb')
    self.write_str(self.header + 't')
    self.write_u32(self.version)
    self.write_u32(self.dim)

    self.write_u32(self.num_means)

    for n in range(self.num_means):
      self.write_u32(self.dim)
      arr_f = array.array('d')
      arr_f.fromlist(list(self.means[n, :]))
      arr_f.tofile(self.f)
      self.write_f64(self.mean_weights[n])

    self.write_u32(self.num_covs)

    for n in range(self.num_covs):
      self.write_u32(self.dim)
      arr_f = array.array('d')
      arr_f.fromlist(list(self.covs[n, :]))
      arr_f.tofile(self.f)
      self.write_f64(self.cov_weights[n])

    self.write_u32(self.num_densities)
    for n in range(self.num_densities):
      self.write_u32(int(self.densities[n, 0]))
      self.write_u32(int(self.densities[n, 1]))

    self.write_u32(self.num_mixtures)
    for n in range(self.num_mixtures):
      num_densities = len(self.mixtures[n][0])
      self.write_u32(num_densities)
      for i in range(num_densities):
        self.write_u32(self.mixtures[n][0][i])
        self.write_f64(self.mixtures[n][1][i])

  def __del__(self):
    self.f.close()

  def get_mean_by_idx(self, idx):
    """
    :param int idx:
    :rtype: numpy.ndarray
    """
    return numpy.float32(self.means[idx, :] / self.mean_weights[idx])

  def get_cov_by_idx(self, idx):
    """
    :param int idx:
    :rtype: numpy.ndarray
    """
    return numpy.float32(self.covs[idx, :] / self.cov_weights[idx])

  def get_number_mixtures(self):
    """
    :rtype: int
    """
    return self.num_mixtures


class WordBoundaries:
  """
  Word boundaries.
  """

  # read routines
  def read_u16(self):
    """
    :rtype: int
    """
    return int(unpack("H", self.f.read(2))[0])

  def read_u32(self):
    """
    :rtype: int
    """
    return int(unpack("i", self.f.read(4))[0])

  def read_str(self, l, enc='ascii'):
    """
    :rtype: str
    """
    a = array.array('b')
    a.fromfile(self.f, l)
    return a.tostring().decode(enc)

  def __init__(self, filename):
    """
    :param str filename:
    """
    self.header = "LATWRDBN"
    self.f = open(filename, 'rb')
    header = self.read_str(8)
    assert header == self.header
    self.version = self.read_u32()
    self.size = self.read_u32()
    print("version=%d size=%d" % (self.version, self.size))
    self.boundaries = []
    for i in range(self.size):
      time = self.read_u32()
      final = self.read_u16()
      initial = self.read_u16()
      bnd = (time, initial, final)
      self.boundaries.append(bnd)
      print(bnd)


###############################################################################

def main():
  """
  Main entry for usage as a tool.
  """
  from argparse import ArgumentParser
  arg_parser = ArgumentParser()
  arg_parser.add_argument("archive")
  arg_parser.add_argument("file", nargs="?")
  arg_parser.add_argument("--mode", default="list", help="list, show")
  arg_parser.add_argument("--type", default="feat", help="ascii, feat, align, bin-matrix, flow-cache")
  arg_parser.add_argument("--allophone-file")
  args = arg_parser.parse_args()

  a = open_file_archive(args.archive, must_exists=True)

  if args.mode == "list":
    print("\n".join(sorted(s for s in a.file_list())))
    sys.exit(0)

  elif args.mode == "show":
    assert args.file, "need to provide 'file' for --mode show. see --help"
    if args.type == "align":
      if args.allophone_file:
        a.set_allophones(args.allophone_file)

      f = a.read(args.file, "align")
      for row in f:
        print(" ".join("%.6f " % x for x in row))

    elif args.type == "feat":
      t, f = a.read(args.file, "feat")
      for row, time in zip(f, t):
        print(str(time) + "--------" + " ".join("%.6f " % x for x in row))

    else:
      raise NotImplementedError("invalid --type %r" % args.type)

  else:
    raise NotImplementedError("invalid --mode %r" % args.mode)


if __name__ == "__main__":
  import better_exchook
  better_exchook.install()
  main()



#!/usr/bin/env python3

import os
import sys
from pprint import pprint
from argparse import ArgumentParser
import tensorflow as tf
from tensorflow.python.client import device_lib

my_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(my_dir))  # parent dir, Returnn code

from TFUtil import get_device_attr, setup_tf_thread_pools, print_available_devices, get_tf_list_local_devices


def dump_devs(tf_session_opts, use_device_lib=False, filter_gpu=True):
  s = os.environ.get("CUDA_VISIBLE_DEVICES", None)
  cuda_num_visible = None
  if s is not None:
    cuda_num_visible = len(s.split(","))
  tf_num_visible = None
  if tf_session_opts.get("gpu_options", {}).get("visible_device_list"):
    tf_num_visible = len(tf_session_opts.get("gpu_options", {}).get("visible_device_list").split(","))
  if use_device_lib:
    devs = list(device_lib.list_local_devices())
  else:
    devs = get_tf_list_local_devices()
  if filter_gpu:
    devs = [dev for dev in devs if dev.device_type == 'GPU']
  # TF, or more likely the CUDA driver, will cache the list_local_devices internally after the first call.
  # Thus even when we update CUDA_VISIBLE_DEVICES, it will not notice that.
  # get_tf_list_local_devices will also cache.
  print("num devs %i, CUDA num visible %r, TF num visible %r" % (len(devs), cuda_num_visible, tf_num_visible))
  print("devs:")
  pprint(devs)
  with tf.Session(config=tf.ConfigProto(**tf_session_opts)) as session:
    for dev in devs:
      print("dev name:", dev.name)
      print("dev attribs:", session.run(get_device_attr(dev.name)))


def main():
  arg_parser = ArgumentParser()
  arg_parser.add_argument("--try_subsets", action="store_true")
  arg_parser.add_argument("--visible_device_list")
  arg_parser.add_argument("--use_device_lib", action="store_true")
  args = arg_parser.parse_args()

  orig_cuda_visible_devs_str = os.environ.get("CUDA_VISIBLE_DEVICES", None)
  print("original CUDA_VISIBLE_DEVICES:", orig_cuda_visible_devs_str)

  tf_session_opts = {}
  if args.visible_device_list:
    tf_session_opts.setdefault("gpu_options", {})["visible_device_list"] = args.visible_device_list
    print("Using TF gpu_options.visible_device_list %r" % args.visible_device_list)

  setup_tf_thread_pools(tf_session_opts=tf_session_opts)
  print_available_devices(tf_session_opts=tf_session_opts)
  dump_devs(tf_session_opts=tf_session_opts, use_device_lib=args.use_device_lib, filter_gpu=False)

  if args.try_subsets:
    print("Trying subsets of CUDA_VISIBLE_DEVICES to see whether list_local_devices is cached.")
    cuda_visible_devs_str = orig_cuda_visible_devs_str
    while cuda_visible_devs_str:
      cuda_visible_devs_str = ",".join(cuda_visible_devs_str.split(",")[:-1])
      print("set CUDA_VISIBLE_DEVICES:", cuda_visible_devs_str)
      os.environ["CUDA_VISIBLE_DEVICES"] = cuda_visible_devs_str
      dump_devs(tf_session_opts=tf_session_opts, use_device_lib=args.use_device_lib)
    os.environ["CUDA_VISIBLE_DEVICES"] = orig_cuda_visible_devs_str
    print("Recovered original CUDA_VISIBLE_DEVICES")
    dump_devs(tf_session_opts=tf_session_opts, use_device_lib=args.use_device_lib)

  print("Quit.")


if __name__ == "__main__":
  import better_exchook
  better_exchook.install()
  main()
